# -*- coding: ascii -*-
"""
ldap0.filters - misc stuff for handling LDAP filter strings (see RFC 4515)
"""

import time
from typing import List, Sequence, Union

from .functions import strf_secs

FILTER_SPECIAL_CHARS = set('\\*()')

# map for escaping special chars in LDAP filters expect back-slash
BASIC_ESCAPE_MAP = {
    ord('*'): '\\2a',
    ord('('): '\\28',
    ord(')'): '\\29',
    ord('\x00'): '\\00',
}


def escape_bytes(assertion_value: bytes) -> str:
    """
    escape all bytes and return str (unicode)
    """
    return ''.join(['\\%02x' % char_code for char_code in assertion_value])


def escape_str(assertion_value: str) -> str:
    """
    Replace all special characters found in assertion_value
    by quoted notation.
    """
    assert isinstance(assertion_value, str), TypeError(
        'Expected assertion_value to be unicode (str), got %r' % (assertion_value,)
    )
    return assertion_value.replace('\\', '\\5c').translate(BASIC_ESCAPE_MAP)


def negate_filter(flt: str) -> str:
    """
    Returns the simple negation of flt
    """
    assert isinstance(flt, str), TypeError(
        'Expected flt to be unicode (str), got %r' % (flt,)
    )
    if flt.startswith('(!') and flt.endswith(')'):
        return flt[2:-1]
    return '(!{0})'.format(flt)


def map_filter_parts(assertion_type: str, assertion_values: Sequence[str]) -> List[str]:
    """
    returns filter parts all with same asserton type but different values
    """
    return [
        '(%s=%s)' % (assertion_type, escape_str(av))
        for av in assertion_values
    ]


def compose_filter(operand: str, filter_parts: Sequence[str]):
    """
    returns filter string composed by operand and list of sub-filters
    """
    if not filter_parts:
        return ''
    if len(filter_parts) == 1:
        return filter_parts[0]
    return '(%s%s)' % (
        operand,
        ''.join(filter_parts),
    )


def dict_filter(
        entry,
        iop: str = '&',
        oop: str = '&',
    ) -> str:
    """
    returns a filter compose from an entry dictionary

    iop
       inner operand used for attribute value list
    oop
       outer operand used for all attributes
    """
    filter_parts = []
    for attr_type, attr_values in entry.items():
        filter_parts.append(
            compose_filter(
                iop,
                map_filter_parts(attr_type, attr_values)
            )
        )
    if len(filter_parts) == 1:
        return filter_parts[0]
    return compose_filter(oop, filter_parts)


def time_span_filter(
        filterstr: str = '',
        from_timestamp: Union[int, float] = 0,
        until_timestamp: Union[int, float, None] = None,
        delta_attr: str = 'modifyTimestamp',
    ):
    """
    If last_run_timestr is non-zero filterstr will be extended
    """
    if until_timestamp is None:
        until_timestamp = time.time()
        if from_timestamp < 0:
            from_timestamp = until_timestamp + from_timestamp
    if from_timestamp > until_timestamp:
        raise ValueError('from_timestamp %r must not be greater than until_timestamp %r' % (
            from_timestamp, until_timestamp
        ))
    return (
        '(&'
        '{filterstr}'
        '({delta_attr}>={from_timestr})'
        '(!({delta_attr}>={until_timestr}))'
        ')'
    ).format(
        filterstr=filterstr,
        delta_attr=delta_attr,
        from_timestr=strf_secs(from_timestamp),
        until_timestr=strf_secs(until_timestamp),
    )
    # end of time_span_filter()



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.base
"""

# from Python's standard lib
import unittest

from ldap0.dn import DNObj
from ldap0.base import decode_entry_dict
from ldap0.res import SearchReference, SearchResultEntry, LDAPResult
from ldap0.ldapurl import LDAPUrl


class Test001SearchReference(unittest.TestCase):
    """
    tests for class ldap0.res.SearchReference
    """

    def test001_simple(self):
        ref_urls = ['ldap://ldap.example.com/dc=example,dc=com']
        sere = SearchReference([ref_urls[0].encode('ascii')], [])
        self.assertEqual(sere.ctrls, [])
        self.assertEqual(sere.ref_url_strings, ref_urls)
        self.assertEqual(sere.ref_ldap_urls, [LDAPUrl(ref_urls[0])])
        sere = SearchReference([ref_urls[0].encode('ascii')], [], encoding='ascii')
        self.assertEqual(sere.ctrls, [])
        self.assertEqual(sere.ref_url_strings, ref_urls)
        self.assertEqual(sere.ref_ldap_urls, [LDAPUrl(ref_urls[0])])

    def test002_unicode(self):
        # default encoding UTF-8
        sere = SearchReference(['ldap://ldap.example.com/o=äöüÄÖÜß'.encode('utf-8')], [])
        self.assertEqual(sere.ref_url_strings,['ldap://ldap.example.com/o=äöüÄÖÜß'])
        self.assertEqual(sere.ref_ldap_urls, [LDAPUrl(sere.ref_url_strings[0])])
        # explicit encoding ISO-8859-1
        sere = SearchReference(['ldap://ldap.example.com/o=ÄÖÜßäöü'.encode('iso-8859-1')], [], encoding='iso-8859-1')
        self.assertEqual(sere.ref_url_strings,['ldap://ldap.example.com/o=ÄÖÜßäöü'])
        self.assertEqual(sere.ref_ldap_urls, [LDAPUrl(sere.ref_url_strings[0])])


    def test003_unicode_errors(self):
        sere = SearchReference(['ldap://ldap.example.com/o=äöüÄÖÜß'.encode('utf-8')], [], encoding='ascii')
        with self.assertRaises(UnicodeDecodeError):
            sere.ref_url_strings
        sere = SearchReference(['ldap://ldap.example.com/o=äöüÄÖÜß'.encode('iso-8859-1')], [])
        with self.assertRaises(UnicodeDecodeError):
            sere.ref_url_strings


class Test002SearchResultEntry(unittest.TestCase):
    """
    tests for class ldap0.res.SearchResultEntry
    """

    def test001_simple(self):
        dn_b = 'o=äöüÄÖÜß,dc=example,dc=com'.encode('utf-8')
        entry_b = {
            b'objectClass': [b'organization'],
            b'o': ['äöüÄÖÜß'.encode('utf-8')],
            b'description': ['ÄÖÜßäöü'.encode('utf-8')],
        }
        sere = SearchResultEntry(dn_b, entry_b, [])
        self.assertEqual(sere.dn_b, dn_b)
        self.assertEqual(sere.entry_b, entry_b)
        self.assertEqual(sere.dn_s, dn_b.decode('utf-8'))
        self.assertEqual(sere.entry_as, {at.decode('ascii'): avs for at, avs in entry_b.items()})
        self.assertEqual(sere.entry_s, decode_entry_dict(entry_b, encoding='utf-8'))
        self.assertEqual(sere.dn_o, DNObj.from_str(dn_b.decode('utf-8')))
        sere = SearchResultEntry('o=äöüÄÖÜß,dc=example,dc=com'.encode('iso-8859-1'), {}, [], encoding='iso-8859-1')
        self.assertEqual(sere.dn_s, sere.dn_b.decode('iso-8859-1'))

    def test002_dn_unicode_error(self):
        sere = SearchResultEntry(
            'o=äöüÄÖÜß,dc=example,dc=com'.encode('iso-8859-1'),
            {
                b'objectClass': [b'organization'],
                b'o': ['äöüÄÖÜß'.encode('iso-8859-1')],
                b'description': ['ÄÖÜßäöü'.encode('iso-8859-1')],
            },
            []
        )
        with self.assertRaises(UnicodeDecodeError):
            sere.dn_s
        with self.assertRaises(UnicodeDecodeError):
            sere.entry_s


class Test004LDAPResult(unittest.TestCase):
    """
    tests for class ldap0.res.LDAPResult
    """

    def test001_simple(self):
        r = LDAPResult(None, None, None, None)
        self.assertEqual(r.rtype, None)
        self.assertEqual(r.rdata, [])
        self.assertEqual(r.msgid, None)
        self.assertEqual(r.ctrls, [])

    def test002_repr(self):
        self.assertEqual(
            repr(LDAPResult(None, None, None, None)),
            'LDAPResult(None, [], None, [])',
        )
        self.assertEqual(
            repr(LDAPResult(None, [], None, [])),
            'LDAPResult(None, [], None, [])',
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Automatic tests for class ldap0.schema.SubSchema
"""

import unittest
from pprint import pformat

import ldap0.ldif
from ldap0.schema.util import urlfetch, modify_modlist
from ldap0.schema.subentry import SubSchema, SCHEMA_ATTR_MAPPING, SCHEMA_CLASS_MAPPING, OIDNotUnique, NameNotUnique
from ldap0.schema.models import AttributeType, ObjectClass, NameForm, Entry, SchemaElementOIDSet

class TestBasics(unittest.TestCase):

    def test_constants(self):
        self.assertEqual(len(SCHEMA_ATTR_MAPPING), 8)
        self.assertEqual(len(SCHEMA_CLASS_MAPPING), 8)


class SubschemaTest(unittest.TestCase):
    """
    Base class for implementing checks based on local schema files
    """
    maxDiff = 10000
    schema_file = 'file:tests/ldif/subschema-openldap-all.ldif'

    @classmethod
    def setUpClass(cls):
        super(SubschemaTest, cls).setUpClass()
        cls.schema_dn, cls.sub_schema = urlfetch(cls.schema_file, check_uniqueness=2)


class TestSubschema(SubschemaTest):
    """
    test ldap0.schema.SubSchema with subschema subentries read from LDIF files
    """

    def test001_basics(self):
        self.assertEqual(self.schema_dn, b'cn=Subschema')
        self.assertEqual(
            len(SCHEMA_CLASS_MAPPING),
            len(self.sub_schema.name2oid),
        )
        self.assertEqual(
            len(SCHEMA_CLASS_MAPPING),
            len(self.sub_schema.sed),
        )

    def test002_attribute_cn(self):
        # check the attribute type 'cn' alias 'commonName'
        oid = self.sub_schema.name2oid[ldap0.schema.models.AttributeType]['cn']
        self.assertEqual(oid, '2.5.4.3')
        self.assertEqual(oid, self.sub_schema.name2oid[ldap0.schema.models.AttributeType]['commonName'])
        cn_at = self.sub_schema.sed[ldap0.schema.models.AttributeType][oid]
        self.assertEqual(cn_at.names, ('cn', 'commonName'))
        self.assertEqual(cn_at.desc, 'RFC4519: common name(s) for which the entity is known by')
        self.assertEqual(cn_at.sup, ('name',))
        self.assertEqual(cn_at.syntax, None)
        self.assertEqual(cn_at.equality, None)
        self.assertEqual(cn_at.substr, None)
        self.assertEqual(cn_at.ordering, None)

    def test003_objectclass_inetorgperson(self):
        # check the object class 'inetOrgPerson'
        oid = self.sub_schema.name2oid[ldap0.schema.models.ObjectClass]['inetOrgPerson']
        self.assertEqual(oid, '2.16.840.1.113730.3.2.2')
        inop_oc = self.sub_schema.sed[ldap0.schema.models.ObjectClass][oid]
        self.assertEqual(inop_oc.names, ('inetOrgPerson',))
        self.assertEqual(inop_oc.desc, 'RFC2798: Internet Organizational Person')
        self.assertEqual(inop_oc.sup, ('organizationalPerson',))
        self.assertEqual(inop_oc.must, ())
        self.assertEqual(
            inop_oc.may,
            (
                'audio', 'businessCategory', 'carLicense',
                'departmentNumber', 'displayName', 'employeeNumber',
                'employeeType', 'givenName', 'homePhone',
                'homePostalAddress', 'initials', 'jpegPhoto',
                'labeledURI', 'mail', 'manager', 'mobile', 'o',
                'pager', 'photo', 'roomNumber', 'secretary', 'uid',
                'userCertificate', 'x500uniqueIdentifier',
                'preferredLanguage', 'userSMIMECertificate',
                'userPKCS12'
            ),
        )

    def test004_attribute_inheritance(self):
        # check SubSchema.get_syntax()
        self.assertEqual(self.sub_schema.get_syntax('does_not_exist_in_self.sub_schema'), None)
        self.assertEqual(self.sub_schema.get_syntax('cn'), '1.3.6.1.4.1.1466.115.121.1.15')
        # check some class attributes inherited by attribute type 'name'
        self.assertEqual(
            self.sub_schema.get_inheritedattr(AttributeType, 'cn', 'syntax'),
            '1.3.6.1.4.1.1466.115.121.1.15',
        )
        self.assertEqual(
            self.sub_schema.get_inheritedattr(AttributeType, 'cn', 'equality'),
            'caseIgnoreMatch',
        )
        self.assertEqual(
            self.sub_schema.get_inheritedattr(AttributeType, 'cn', 'substr'),
            'caseIgnoreSubstringsMatch',
        )
        cn_i_at = self.sub_schema.get_inheritedobj(
            AttributeType,
            'cn',
            ('syntax', 'equality', 'substr', 'ordering'),
        )
        self.assertEqual(cn_i_at.sup, ('name',))
        self.assertEqual(cn_i_at.syntax, '1.3.6.1.4.1.1466.115.121.1.15')
        self.assertEqual(cn_i_at.equality, 'caseIgnoreMatch')
        self.assertEqual(cn_i_at.substr, 'caseIgnoreSubstringsMatch')
        self.assertEqual(cn_i_at.ordering, None)

    def test005_determine_no_user_mod_attrs(self):
        """
        test SubSchema.determine_no_user_mod_attrs()
        """
        no_user_mod_attrs = self.sub_schema.determine_no_user_mod_attrs()
        self.assertEqual(len(no_user_mod_attrs), 54)
        self.assertIn('2.5.18.4', no_user_mod_attrs) # modifiersName
        no_user_mod_attrs_set = SchemaElementOIDSet(self.sub_schema, AttributeType, no_user_mod_attrs)
        self.assertIn('2.5.18.4', no_user_mod_attrs_set)
        self.assertIn('modifiersName', no_user_mod_attrs_set)
        # test SubSchema.determine_no_user_mod_attrs()
        op_attrs = self.sub_schema.get_all_operational_attributes()
        self.assertIn('2.5.18.4', op_attrs)
        self.assertIn('modifiersName', op_attrs)
        # check SubSchema.get_applicable_aux_classes()
        aux_classes = self.sub_schema.get_applicable_aux_classes('aeUser-dcr')
        self.assertEqual(
            sorted(aux_classes),
            [
                'inetLocalMailRecipient', 'ldapPublicKey', 'msPwdResetObject',
                'oathHOTPUser', 'oathTOTPUser', 'pkiUser', 'posixAccount'
            ],
        )

    def test006_get_structural_oc(self):
        """
        check SubSchema.get_structural_oc()
        """
        self.assertEqual(self.sub_schema.get_structural_oc([]), None)
        self.assertEqual(self.sub_schema.get_structural_oc(['posixAccount']), None)
        self.assertEqual(
            self.sub_schema.get_structural_oc(['inetOrgPerson']),
            '2.16.840.1.113730.3.2.2',
        )
        self.assertEqual(
            self.sub_schema.get_structural_oc(['inetOrgPerson', 'posixAccount']),
            '2.16.840.1.113730.3.2.2',
        )
        self.assertEqual(
            self.sub_schema.get_structural_oc([
                'person', 'organizationalPerson', 'inetOrgPerson', 'posixAccount'
            ]),
            '2.16.840.1.113730.3.2.2',
        )

    def test007_models_entry(self):
        """
        test ldap0.schema.models.Entry
        """
        ldap_entry = {
            b'objectClass': [b'aeSrvGroup'],
            b'cn': [b'test-srvgroup-1'],
            b'description': [b'testing...'],
        }
        ae_dn = 'cn=test-srvgroup-1,ou=ae-dir'
        ae_entry = Entry(self.sub_schema, ae_dn, ldap_entry)
        self.assertIn('cn', ae_entry)
        self.assertIn('2.5.4.3', ae_entry)
        self.assertNotIn('aeLoginGroups', ae_entry)
        self.assertEqual(
            list(ae_entry.get_possible_dit_structure_rules(ae_dn)),
            ['5'],
        )
        self.assertEqual(
            sorted(ae_entry.keys()),
            sorted([at.decode('ascii') for at in ldap_entry.keys()])
        )
        self.assertEqual(
            sorted(ae_entry.items()),
            sorted([(at.decode('ascii'), av) for at, av in ldap_entry.items()])
        )
        self.assertEqual(
            sorted(ae_entry.object_class_oid_set()),
            ['1.3.6.1.4.1.5427.1.389.100.6.13'],
        )
        self.assertEqual(ae_entry.get_rdn_templates(), ['cn='])

    def test008_name_forms(self):
        """
        test SubSchema.get_applicable_name_form_objs()
        """
        aeperson_oid = self.sub_schema.get_oid(ObjectClass, 'aePerson')
        self.assertEqual(aeperson_oid, '1.3.6.1.4.1.5427.1.389.100.6.8')
        name_forms = self.sub_schema.get_applicable_name_form_objs(
            'uniqueIdentifier=foo,cn=hr,ou=ae-dir',
            aeperson_oid,
        )
        self.assertEqual(len(name_forms), 1)
        self.assertEqual(
            str(name_forms[0]),
            str(NameForm("( 1.3.6.1.4.1.5427.1.389.42.15.12 NAME 'aePerson-aeZone-NF-uniqueIdentifier' OC aePerson MUST ( uniqueIdentifier ) )")),
        )

    def test009_tree(self):
        """
        test SubSchema.tree()
        """
        tree = self.sub_schema.tree(AttributeType)
        self.assertEqual(len(tree), 1490)

    def test010_check_uniqueness(self):
        """
        provoke exceptions to be raised because of non-unique OIDs or NAMEs
        """
        for non_unique_oid_tests in (
            {
                'attributeTypes': [
                    '( 1.2.3.4 NAME foo )',
                    '( 1.2.3.4 NAME bar )',
                ],
            },
            {
                'objectClasses': [
                    '( 1.2.3.4 NAME foo )',
                    '( 1.2.3.4 NAME bar )',
                ],
            },
        ):
            with self.assertRaises(OIDNotUnique):
                SubSchema(non_unique_oid_tests, check_uniqueness=2)
        for non_unique_name_tests in (
            {
                'attributeTypes': [
                    '( 1.2.3.4 NAME foo )',
                    '( 2.3.4.5 NAME foo )',
                ],
            },
            {
                'objectClasses': [
                    '( 1.2.3.4 NAME foo )',
                    '( 2.3.4.5 NAME foo )',
                ],
            },
        ):
            with self.assertRaises(NameNotUnique):
                SubSchema(non_unique_name_tests, check_uniqueness=2)

    def test011_get_subord_structural_oc_names(self):
        self.assertEqual(
            list(self.sub_schema.get_subord_structural_oc_names('42')[0]),
            [],
        )


class TestSchemaElementOIDSet(SubschemaTest):
    """
    Tests for class SchemaElementOIDSet
    """

    def test001_AttributeType(self):
        self.assertNotIn('cn', SchemaElementOIDSet(self.sub_schema, AttributeType, []))
        self.assertNotIn('cn', SchemaElementOIDSet(self.sub_schema, AttributeType, ['objectClass']))
        self.assertIn('cn', SchemaElementOIDSet(self.sub_schema, AttributeType, ['cn']))
        self.assertIn('2.5.4.3', SchemaElementOIDSet(self.sub_schema, AttributeType, ['cn']))
        self.assertIn('cn', SchemaElementOIDSet(self.sub_schema, AttributeType, ['2.5.4.3']))
        self.assertIn('commonName', SchemaElementOIDSet(self.sub_schema, AttributeType, ['2.5.4.3']))
        self.assertEqual(len(SchemaElementOIDSet(self.sub_schema, AttributeType, ['2.5.4.3', 'cn', 'commonName'])), 1)
        self.assertEqual(len(SchemaElementOIDSet(self.sub_schema, AttributeType, ['', ' ', '  '])), 0)
        self.assertIn('cn', SchemaElementOIDSet(self.sub_schema, AttributeType, ['@person']))
        self.assertEqual(
            sorted(SchemaElementOIDSet(self.sub_schema, AttributeType, ['@person'])),
            [
                '2.5.4.0',
                '2.5.4.13',
                '2.5.4.20',
                '2.5.4.3',
                '2.5.4.34',
                '2.5.4.35',
                '2.5.4.4',
            ]
        )
        self.assertEqual(
            sorted(SchemaElementOIDSet(self.sub_schema, AttributeType, ['foo', 'bar'])),
            ['bar', 'foo']
        )
        with self.assertRaises(KeyError):
            SchemaElementOIDSet(self.sub_schema, AttributeType, ['@foo'])
        se_set = SchemaElementOIDSet(self.sub_schema, AttributeType, ['cn'])
        self.assertNotIn('mail', se_set)
        se_set.add('mail')
        self.assertIn('mail', se_set)

    def test002_ObjectClass(self):
        self.assertNotIn('person', SchemaElementOIDSet(self.sub_schema, ObjectClass, []))
        self.assertIn('person', SchemaElementOIDSet(self.sub_schema, ObjectClass, ['person']))
        self.assertIn('person', SchemaElementOIDSet(self.sub_schema, ObjectClass, ['2.5.6.6']))
        self.assertEqual(len(SchemaElementOIDSet(self.sub_schema, ObjectClass, ['2.5.6.6', 'person'])), 1)
        self.assertEqual(len(SchemaElementOIDSet(self.sub_schema, ObjectClass, ['', ' ', '  '])), 0)
        self.assertEqual(
            sorted(SchemaElementOIDSet(self.sub_schema, ObjectClass, ['foo', 'bar'])),
            ['bar', 'foo']
        )
        with self.assertRaises(ValueError):
            SchemaElementOIDSet(self.sub_schema, ObjectClass, ['@foo'])
        se_set = SchemaElementOIDSet(self.sub_schema, ObjectClass, ['person'])
        self.assertNotIn('organizationalPerson', se_set)
        se_set.add('organizationalPerson')
        self.assertIn('organizationalPerson', se_set)

    def test003_names_property(self):
        self.assertEqual(SchemaElementOIDSet(self.sub_schema, ObjectClass, ['person', '2.5.6.6']).names, ['person'])
        self.assertEqual(SchemaElementOIDSet(self.sub_schema, AttributeType, ['cn', 'commonName', '2.5.4.3']).names, ['cn'])
        self.assertEqual(
            sorted(SchemaElementOIDSet(self.sub_schema, AttributeType, ['@person']).names),
            [
                'cn',
                'description',
                'objectClass',
                'seeAlso',
                'sn',
                'telephoneNumber',
                'userPassword',
            ]
        )


class TestModifyModlist(SubschemaTest):
    modify_modlist_tests = [
        (
            {
                b'objectClass':[b'person', b'pilotPerson'],
                b'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                b'sn':[b'Str\303\266der'],
                b'enum':[b'a', b'b', b'c'],
                b'c':[b'DE'],
            },
            {
                b'objectClass':[b'person', b'inetOrgPerson'],
                b'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                b'sn':[],
                b'enum':[b'a', b'b', b'd'],
                b'mail':[b'michael@stroeder.com'],
            },
            [],
            [
                (ldap0.MOD_DELETE, b'objectClass', [b'person', b'pilotPerson']),
                (ldap0.MOD_ADD, b'objectClass', [b'person', b'inetOrgPerson']),
                (ldap0.MOD_DELETE, b'c', [b'DE']),
                (ldap0.MOD_DELETE, b'sn', [b'Str\303\266der']),
                (ldap0.MOD_ADD, b'mail', [b'michael@stroeder.com']),
                (ldap0.MOD_DELETE, b'enum', None),
                (ldap0.MOD_ADD, b'enum', [b'a', b'b', b'd']),
            ]
        ),

        (
            {
                b'c':[b'DE'],
            },
            {
                b'c':[b'FR'],
            },
            [],
            [
                (ldap0.MOD_DELETE, b'c', [b'DE']),
                (ldap0.MOD_ADD, b'c', [b'FR']),
            ]
        ),

        (
            {
                b'objectClass':[b'person'],
                b'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                b'sn':[b'Str\303\266der'],
                b'enum':[b'a', b'b'],
            },
            {
                b'objectClass':[b'inetOrgPerson'],
                b'cn':[b'Michael Str\303\266der', b'Michael Stroeder'],
                b'sn':[],
                b'enum':[b'a', b'B'],
            },
            ['objectClass'],
            [
                (ldap0.MOD_DELETE, b'sn', [b'Str\303\266der']),
                (ldap0.MOD_DELETE, b'enum', None),
                (ldap0.MOD_ADD, b'enum', [b'a', b'B']),
            ]
        ),

    ]

    def test_modify_modlist(self):
        for old_entry, new_entry, ignore_attr_types, test_modlist in self.modify_modlist_tests:
            result_modlist = modify_modlist(
                self.sub_schema,
                ldap0.schema.models.Entry(self.sub_schema, 'cn=test', old_entry),
                ldap0.schema.models.Entry(self.sub_schema, 'cn=test', new_entry),
                ignore_attr_types=ignore_attr_types,
            )
            self.assertEqual(
                sorted(test_modlist),
                sorted(result_modlist),
                '\nmodify_modlist(\n%s,\n%s\n)\nreturns\n%s\ninstead of\n%s.' % (
                    pformat(old_entry),
                    pformat(new_entry),
                    pformat(sorted(result_modlist)),
                    pformat(sorted(test_modlist)),
                )
            )
            result_modlist = modify_modlist(
                self.sub_schema,
                ldap0.schema.models.Entry(self.sub_schema, 'cn=test', old_entry),
                ldap0.schema.models.Entry(self.sub_schema, 'cn=test', new_entry),
                ignore_attr_types=SchemaElementOIDSet(self.sub_schema, AttributeType, ignore_attr_types),
            )
            self.assertEqual(
                sorted(test_modlist),
                sorted(result_modlist),
                '\nmodify_modlist(\n%s,\n%s\n)\nreturns\n%s\ninstead of\n%s.' % (
                    pformat(old_entry),
                    pformat(new_entry),
                    pformat(sorted(result_modlist)),
                    pformat(sorted(test_modlist)),
                )
            )


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.res - basic classes for LDAP results
"""

from typing import Dict, List, Optional, Union, Sequence, Tuple

from pyasn1.error import PyAsn1Error

import _libldap0

from .dn import DNObj
from .base import decode_list
from .ldapurl import LDAPUrl
from .typehints import EntryBytes, EntryMixed, EntryStr
from .controls import ResponseControl, RequestControl, KNOWN_RESPONSE_CONTROLS
from .extop import ExtendedResponse

__all__ = [
    'LDAPResult',
    'SearchReference',
    'SearchResultEntry',
    'encode_request_ctrls',
    'decode_response_ctrls',
]


class IntermediateResponse:
    """
    Generic base class for a LDAPv3 intermediate response

    requestName
        OID as string of the LDAPv3 intermediate response
    encodedResponseValue
        BER-encoded ASN.1 value of the LDAPv3 intermediate response
    """
    encoding = 'utf-8'

    responseName = None

    def __init__(self, responseName=None, encodedResponseValue=None, ctrls=None):
        if responseName is not None:
            self.responseName = responseName
        self.responseValue = encodedResponseValue
        if encodedResponseValue is not None:
            self.decode(encodedResponseValue)
        self.ctrls = ctrls or []

    def __repr__(self):
        return '%s(responseName=%r, responseValue=%r)' % (
            self.__class__.__name__,
            self.responseName,
            self.responseValue,
        )

    @classmethod
    def check_resp_name(cls, name):
        """
        returns True if :name: is the correct expected responseName
        """
        return name == cls.responseName

    def decode(self, value: bytes):
        """
        decodes the BER-encoded ASN.1 extended operation response value and
        sets the appropriate class attributes
        """
        self.responseValue = value


class IntermediateResponseRegistry:
    """
    A simple registry for responses and their handler class
    """

    _handler_cls: Dict[str, Union[ExtendedResponse, IntermediateResponse]]

    def __init__(self):
        self._handler_cls = {}

    def register(
            self,
            handler: Union[ExtendedResponse, IntermediateResponse],
        ):
        """
        register an handler class for extended/intermediate response
        """
        self._handler_cls[handler.responseName] = handler

    def get(
            self,
            oid: str,
            default: Optional[Union[ExtendedResponse, IntermediateResponse]] = None,
        ) -> Optional[Union[ExtendedResponse, IntermediateResponse]]:
        """
        return handler class for extended/intermediate response by OID (responseName)
        """
        return self._handler_cls.get(oid, default)


# response OID to class registry
INTERMEDIATE_RESPONSE_REGISTRY = IntermediateResponseRegistry()


class SearchReference:
    """
    LDAP search continuation objects
    """
    __slots__ = (
        'ctrls',
        'encoding',
        '_ref_urls_b',
        '_ref_urls_o',
        '_ref_urls_s',
    )
    _ref_urls_s: Optional[List[str]]
    _ref_urls_o: Optional[List[LDAPUrl]]

    def __init__(self, entry: List[bytes], ctrls=None, encoding: str = 'utf-8'):
        self.encoding = encoding
        self._ref_urls_b = entry
        self._ref_urls_s = self._ref_urls_o = None
        self.ctrls = ctrls

    def __repr__(self) -> str:
        return '%s(%r, %r, encoding=%r)' % (
            self.__class__.__name__,
            self._ref_urls_b,
            self.ctrls,
            self.encoding,
        )

    @property
    def ref_url_strings(self) -> List[str]:
        if self._ref_urls_s is None:
            self._ref_urls_s = decode_list(self._ref_urls_b, encoding=self.encoding)
        return self._ref_urls_s

    @property
    def ref_ldap_urls(self) -> List[LDAPUrl]:
        if self._ref_urls_o is None:
            self._ref_urls_o = [LDAPUrl(url) for url in self.ref_url_strings]
        return self._ref_urls_o


class SearchResultEntry:
    """
    LDAP search result objects
    """
    __slots__ = (
        'ctrls',
        'dn_b',
        '_dn_o',
        '_dn_s',
        'encoding',
        '_entry_as',
        '_entry_b',
        '_entry_s',
    )
    dn_b: bytes
    _dn_s: Optional[str]
    _dn_o: Optional[DNObj]
    _entry_b: EntryBytes
    _entry_as: Optional[EntryMixed]
    _entry_s: Optional[EntryStr]

    def __init__(self, dn: bytes, entry: EntryBytes, ctrls=None, encoding: str = 'utf-8'):
        self.encoding = encoding
        self.dn_b = dn
        self._dn_s = self._dn_o = None
        self._entry_b = entry
        self._entry_as = self._entry_s = None
        self.ctrls = ctrls or []

    def __repr__(self) -> str:
        return '%s(%r, %r, %r, encoding=%r)' % (
            self.__class__.__name__,
            self.dn_b,
            self._entry_b,
            self.ctrls,
            self.encoding,
        )

    def __eq__(self, other):
        return (
            self.dn_s == other.dn_s
            and self.entry_as == other.entry_as
            and self.ctrls == other.ctrls
        )

    @property
    def dn_s(self):
        if self._dn_s is None:
            self._dn_s = self.dn_b.decode(self.encoding)
        return self._dn_s

    @property
    def dn_o(self):
        if self._dn_o is None:
            self._dn_o = DNObj.from_str(self.dn_s)
        return self._dn_o

    @property
    def entry_b(self) -> EntryBytes:
        return self._entry_b

    @property
    def entry_as(self) -> EntryMixed:
        if self._entry_as is None:
            self._entry_as: EntryMixed = {
                at.decode('ascii'): avs
                for at, avs in self._entry_b.items()
            }
        return self._entry_as

    @property
    def entry_s(self) -> EntryStr:
        if self._entry_s is None:
            self._entry_s: EntryStr = {
                at.decode('ascii'): [av.decode(self.encoding) for av in avs]
                for at, avs in self._entry_b.items()
            }
        return self._entry_s


class LDAPResult:
    """
    base class for LDAP result objects
    """
    __slots__ = (
        'rtype',
        'rdata',
        'msgid',
        'ctrls',
        'call_args',
    )
    rtype: int
    rdata: List[Union[SearchReference, SearchResultEntry, IntermediateResponse]]
    msgid: int
    ctrls: List[ResponseControl]

    def __init__(
            self,
            rtype: int,
            rdata,
            msgid: int,
            ctrls,
            encoding: str = 'utf-8',
            call_args = None,
        ):
        self.rtype = rtype
        self.rdata = []
        self.msgid = msgid
        self.ctrls = ctrls or []
        self.call_args = call_args
        for dn, entry, ctrls in rdata or []:
            if rtype == _libldap0.RES_SEARCH_REFERENCE:
                self.rdata.append(SearchReference(entry, ctrls, encoding=encoding))
            elif rtype == _libldap0.RES_SEARCH_ENTRY:
                self.rdata.append(SearchResultEntry(dn, entry, ctrls, encoding=encoding))
            elif rtype == _libldap0.RES_INTERMEDIATE:
                respoid = dn.decode(encoding)
                resp_class = INTERMEDIATE_RESPONSE_REGISTRY.get(respoid, IntermediateResponse)
                response = resp_class(respoid, encodedResponseValue=entry, ctrls=ctrls)
                self.rdata.append(response)
            else:
                raise NotImplementedError

    def __repr__(self) -> str:
        return '%s(%r, %r, %r, %r)' % (
            self.__class__.__name__,
            self.rtype,
            self.rdata,
            self.msgid,
            self.ctrls,
        )


def encode_request_ctrls(
        ctrls: Sequence[RequestControl]
    ) -> Tuple[str, bool, bytes]:
    """
    Return list of readily encoded 3-tuples which can be directly
    passed to C module _ldap

    ctrls
        sequence-type of RequestControl objects
    """
    if ctrls is None:
        return None
    return [
        (c.controlType.encode('ascii'), c.criticality, c.encode())
        for c in ctrls
    ]
    # end of encode_request_ctrls()


def decode_response_ctrls(
        ctrls: Sequence[ResponseControl],
        ctrl_reg: Optional[Dict[str, ResponseControl]] = None,
    ):
    """
    Returns list of readily decoded ResponseControl objects

    ctrls
        Sequence-type of 3-tuples returned by _libldap0.result() containing
        the encoded ASN.1 control values of response controls.
    ctrl_reg
        Dictionary mapping extended control's OID to ResponseControl class
        of response controls known by the application. If None
        ldap0.controls.KNOWN_RESPONSE_CONTROLS is used here.
    """
    if ctrl_reg is None:
        ctrl_reg = KNOWN_RESPONSE_CONTROLS
    result = []
    for ctrl_type, ctrl_criticality, ctrl_val in ctrls or []:
        ctrl_type = ctrl_type.decode('ascii')
        try:
            control_class = ctrl_reg.get(ctrl_type)
        except KeyError:
            if ctrl_criticality:
                raise _libldap0.UNAVAILABLE_CRITICAL_EXTENSION(
                    'Unknown critical response control, ctrl_type %r' % (ctrl_type)
                )
        else:
            control = control_class(criticality=ctrl_criticality)
            try:
                control.decode(ctrl_val)
            except PyAsn1Error as err:
                if ctrl_criticality:
                    raise err
            else:
                result.append(control)
    return result
    # end of decode_response_ctrls()



# -*- coding: ascii -*-
"""
ldap0.controls - support classes for LDAP controls

The ldap0.controls module provides LDAPControl classes.
Each class provides support for a certain control.
"""

import inspect
from importlib import import_module
from typing import Dict, Optional, Sequence, Tuple

import _libldap0


__all__ = [
    # Classes
    'LDAPControl',
    'RequestControl',
    'ResponseControl',
]


class RequestControl:
    """
    Base class for all request controls

    controlType
        OID as string of the LDAPv3 extended request control
    criticality
        sets the criticality of the control (boolean)
    encodedControlValue
        control value of the LDAPv3 extended request control
        (here it is the BER-encoded ASN.1 control value)
    """
    encoding: str = 'utf-8'
    controlType: Optional[str] = None

    def __init__(
            self,
            controlType: Optional[str] = None,
            criticality: bool = False,
            encodedControlValue: Optional[bytes] = None
        ):
        if controlType is not None:
            self.controlType = controlType
        self.criticality = criticality
        self.encodedControlValue = encodedControlValue

    def encode(self) -> bytes:
        """
        sets class attribute encodedControlValue to the BER-encoded ASN.1
        control value composed by class attributes set before
        """
        return self.encodedControlValue


class ResponseControl:
    """
    Base class for all response controls

    controlType
        OID as string of the LDAPv3 extended response control
    criticality
        sets the criticality of the received control (boolean)
    """
    encoding: str = 'utf-8'

    def __init__(
            self,
            controlType: Optional[str] = None,
            criticality: bool = False,
        ):
        if controlType is not None:
            self.controlType = controlType
        self.criticality = criticality

    def decode(self, encodedControlValue: bytes):
        """
        decodes the BER-encoded ASN.1 control value and sets the appropriate
        class attributes
        """
        self.controlValue = encodedControlValue


class LDAPControl(RequestControl, ResponseControl):
    """
    Base class for combined request/response controls
    """
    encoding: str = 'utf-8'
    controlType: Optional[str] = None

    def __init__(
            self,
            controlType: Optional[str] = None,
            criticality: bool = False,
            encodedControlValue: Optional[bytes] = None
        ):
        RequestControl.__init__(
            self,
            controlType=controlType,
            criticality=criticality,
            encodedControlValue=encodedControlValue,
        )


class ControlRegistry:
    """
    A simple registry for extended controls and their handler class
    """

    cmap: Dict[str, ResponseControl]

    def __init__(self, *module_names):
        self.cmap = None
        self._module_names = module_names

    def _reg(self, *module_names):
        for module_name in module_names:
            module = import_module(module_name)
            for _, cls in inspect.getmembers(module, inspect.isclass):
                if (
                        issubclass(cls, ResponseControl)
                        and hasattr(cls, 'controlType')
                        and cls.controlType is not None
                    ):
                    self.cmap[cls.controlType] = cls

    def get(self, control_type):
        if self.cmap is None:
            self.cmap = {}
            self._reg(*self._module_names)
        return self.cmap[control_type]

# response control OID to class registry
KNOWN_RESPONSE_CONTROLS = ControlRegistry(
    __name__+'.deref',
    __name__+'.libldap',
    __name__+'.openldap',
    __name__+'.pagedresults',
    __name__+'.ppolicy',
    __name__+'.psearch',
    __name__+'.pwdpolicy',
    __name__+'.readentry',
    __name__+'.sessiontrack',
    __name__+'.simple',
    __name__+'.sss',
    __name__+'.syncrepl',
    __name__+'.vlv',
)



# -*- coding: utf-8 -*-
"""
Simple tests for module ldap0.controls
"""

# from Python's standard lib
import unittest

from pyasn1.error import PyAsn1Error

import _libldap0
import ldap0.res
import ldap0.controls
from ldap0.controls import ResponseControl, ControlRegistry
from ldap0.controls.sessiontrack import *
from ldap0.controls.psearch import *
from ldap0.controls.ppolicy import *
from ldap0.controls.pwdpolicy import *
from ldap0.controls.readentry import *
from ldap0.controls.syncrepl import *
from ldap0.controls.vlv import *
from ldap0.controls.libldap import *
from ldap0.controls.deref import *
from ldap0.res import encode_request_ctrls, decode_response_ctrls

# re-raise pyasn1 decoding exceptions
ldap0.controls.STRICT_DECODE = True


class TestRequestControls(unittest.TestCase):
    """
    test classes for encoding request controls
    """

    def test_sessiontrack(self):
        stc = SessionTrackingControl(
            '12.34.56.78',
            'tracking-NAME',
            SESSION_TRACKING_FORMAT_OID_RADIUS_ACCT_SESSION_ID,
            'tracking-ID',
        )
        stc_encoded = b'0G\x04\x0b12.34.56.78\x04\rtracking-NAME\x04\x1c1.3.6.1.4.1.21008.108.63.1.1\x04\x0btracking-ID'
        self.assertEqual(stc.encode(), stc_encoded)
        self.assertEqual(
            encode_request_ctrls([stc]),
            [(
                SessionTrackingControl.controlType.encode('ascii'),
                False,
                stc_encoded,
            )],
        )

    def test_psearch(self):
        psc = PersistentSearchControl(
            criticality=True,
            changeTypes=None,
            changesOnly=False,
            returnECs=True,
        )
        psc_encoded = b'0\t\x02\x01\x0f\x01\x01\x00\x01\x01\x01'
        self.assertEqual(psc.encode(), psc_encoded)
        self.assertEqual(
            encode_request_ctrls([psc]),
            [(
                PersistentSearchControl.controlType.encode('ascii'),
                True,
                psc_encoded,
            )],
        )


class TestResponseControls(unittest.TestCase):
    """
    test classes for decoding response controls
    """

    def test_base(self):
        rc = ResponseControl()
        rc.decode(b'foo')
        self.assertEqual(rc.controlValue, b'foo')
        with self.assertRaises(_libldap0.UNAVAILABLE_CRITICAL_EXTENSION):
            decode_response_ctrls(
                [(b'1.2.3.4', True, b'')],
            )
        with self.assertRaises(ValueError):
            decode_response_ctrls(
                [(PasswordExpiringControl.controlType.encode('ascii'), False, b'')],
            )
        with self.assertRaises(PyAsn1Error):
            decode_response_ctrls(
                [(PasswordPolicyControl.controlType.encode('ascii'), True, b'\xFF\x00\x00\x00\x00\x01')],
            )
        self.assertEqual(
            decode_response_ctrls(
                [(PasswordExpiringControl.controlType.encode('ascii'), False, b'')],
                ctrl_reg=ControlRegistry(),
            ),
            [],
        )

    def test_pwdpolicy(self):
        pec = PasswordExpiringControl()
        pec.decode(b'0')
        self.assertEqual(pec.gracePeriod, 0)
        with self.assertRaises(ValueError):
            pec.decode(b'x')
        pec = PasswordExpiredControl()
        pec.decode(b'0')
        self.assertEqual(pec.passwordExpired, True)
        pec.decode(b'x')
        self.assertEqual(pec.passwordExpired, False)

    def test_readentry(self):
        # test fall-back decoding caused by old OpenLDAP issue ITS#6899
        broken_ber = b'0c\x043cn=Samba Unix UID Pool,ou=Testing,dc=stroeder,dc=de0,0\x14\x04\tuidNumber1\x07\x04\x05100050\x14\x04\tgidNumber1\x07\x04\x0510005'
        rec = ReadEntryControl()
        rec.decode(broken_ber)
        self.assertEqual(rec.res.dn_s, 'cn=Samba Unix UID Pool,ou=Testing,dc=stroeder,dc=de')
        self.assertEqual(rec.res.entry_as, {'gidNumber': [b'10005'], 'uidNumber': [b'10005']})
        with self.assertRaises(PyAsn1Error):
            rec.decode(b'\xFF')


class TestVirtualListView(unittest.TestCase):
    """
    test class for ldap0.controls.vlv
    """

    @unittest.skip('Test vectors not verified')
    def test001_vlv_request_control(self):
        vrc = VLVRequestControl(
            criticality=True,
            before_count=10,
            after_count=20,
            offset=10,
            content_count=10,
            greater_than_or_equal=None,
            context_id=None,
        )
        self.assertEqual(vrc.encode(), b'0\x0e\x02\x01\n\x02\x01\x14\xa0\x06\x02\x01\n\x02\x01\n')
        vrc = VLVRequestControl(
            criticality=True,
            before_count=10,
            after_count=20,
            offset=10,
            content_count=None,
            greater_than_or_equal=10,
            context_id=None,
        )
        self.assertEqual(vrc.encode(), b'0\n\x02\x01\n\x02\x01\x14\x81\x0210')

    def test002_vlv_response_control(self):
        vrc = VLVResponseControl(criticality=True)
        with self.assertRaises(PyAsn1Error):
            vrc.decode(b'\xFF')


class TestAssertionControl(unittest.TestCase):

    def test001_assertion_control(self):
        self.assertEqual(
            AssertionControl(True, '(entryDN=uid=foo,cn=ae,ou=ae-dir)').encode(),
            b'\xa3"\x04\x07entryDN\x04\x17uid=foo,cn=ae,ou=ae-dir'
        )
        self.assertEqual(
            AssertionControl(
                True,
                (
                    '(&(entryCSN=20190917170703.426971Z#000000#000#000000)(entryDN=cn=Schwitters\\5c2C Kurt,ou=Testing,dc=stroeder,dc=de)(entryUUID=58e450f6-b2e0-1038-8849-9fc73698254c)(createTimestamp=20190122222520Z)(modifyTimestamp=20190917170703Z)(creatorsName=cn=michael ströder+mail=michael@stroeder.com,ou=private,dc=stroeder,dc=de)(modifiersName=cn=michael ströder+mail=michael@stroeder.com,ou=private,dc=stroeder,dc=de))'
                )
            ).encode(),
            b'\xa0\x82\x01\xa9\xa34\x04\x08'
            b'entryCSN\x04(20190917170703.426971Z#000000#000#000000'
            b'\xa3=\x04\x07entryDN\x042cn=Schwitters\\2C Kurt,ou=Testing,dc=stroeder,dc=de'
            b'\xa31\x04\tentryUUID\x04$58e450f6-b2e0-1038-8849-9fc73698254c\xa3"'
            b'\x04\x0fcreateTimestamp\x04\x0f20190122222520Z\xa3"'
            b'\x04\x0fmodifyTimestamp\x04\x0f20190917170703Z\xa3Z'
            b'\x04\x0ccreatorsName\x04Jcn=michael str\xc3\xb6der+mail=michael@stroeder.com,ou=private,dc=stroeder,dc=de'
            b'\xa3[\x04\rmodifiersName\x04Jcn=michael str\xc3\xb6der+mail=michael@stroeder.com,ou=private,dc=stroeder,dc=de'
        )


class TestSyncRepl(unittest.TestCase):
    """
    test class for ldap0.controls.syncrepl
    """

    #@unittest.skip('Test vectors not verified')
    def test001_syncrepl_request_control(self):
        src = SyncRequestControl(
            criticality=True,
        )
        self.assertEqual(src.encode(), b'0\x03\n\x01\x01')

        src = SyncRequestControl(
            criticality=True,
            cookie="rid=000,csn=20190917170703.426971Z#000000#000#000000",
            mode='refreshAndPersist',
            reloadHint=False
        )
        self.assertEqual(src.encode(), b'09\n\x01\x03\x044rid=000,csn=20190917170703.426971Z#000000#000#000000')

    def _check_state_decoded_val(self, buf, state=None, entryUUID=None, cookie=None):
        ssc = SyncStateControl()
        ssc.decode(buf)
        self.assertEqual(ssc.state, state)
        self.assertEqual(ssc.entryUUID, entryUUID)
        self.assertEqual(ssc.cookie, cookie)

    def test002_syncrepl_state_response_control(self):
        self._check_state_decoded_val(
            b'0K\x0a\x01\x01\x04\x10\xefg\x1dwU\tJ\xb5\xbb\xd3j\xb8~&\x1b;\x044rid=000,csn=20200414133753.354835Z#000000#000#000000',
            state='add',
            entryUUID='ef671d77-5509-4ab5-bbd3-6ab87e261b3b',
            cookie=b'rid=000,csn=20200414133753.354835Z#000000#000#000000',
        )

    def _check_done_decoded_val(self, buf, cookie=None, refreshDeletes=None):
        sdc = SyncDoneControl()
        sdc.decode(buf)
        self.assertEqual(sdc.cookie, cookie)
        self.assertEqual(sdc.refreshDeletes, refreshDeletes)

    def test003_syncrepl_done_response_control(self):
        self._check_done_decoded_val(
            b'09\x044rid=000,csn=20200414133753.354835Z#000000#000#000000\x01\x01\xff',
            cookie=b'rid=000,csn=20200414133753.354835Z#000000#000#000000',
            refreshDeletes=True,
        )


class TestPPolicy(unittest.TestCase):

    def _check_decoded_val(self, pp_buf, timeBeforeExpiration=None, graceAuthNsRemaining=None, error=None):
        pp = PasswordPolicyControl()
        pp.decode(pp_buf)
        self.assertEqual(pp.timeBeforeExpiration, timeBeforeExpiration)
        self.assertEqual(pp.graceAuthNsRemaining, graceAuthNsRemaining)
        self.assertEqual(pp.error, error)

    def test_ppolicy_graceauth(self):
        self._check_decoded_val(
            b'0\x84\x00\x00\x00\t\xa0\x84\x00\x00\x00\x03\x81\x01\x02',
            graceAuthNsRemaining=2,
        )

    def test_ppolicy_timebefore(self):
        self._check_decoded_val(
            b'0\x84\x00\x00\x00\t\xa0\x84\x00\x00\x00\x03\x80\x012',
            timeBeforeExpiration=50,
        )


class TestDeref(unittest.TestCase):

    def test_deref(self):
        deref = DereferenceControl(criticality=False)
        deref.decode(
            b'0\x81\xae0U\x04\x06member\x04\x1auid=rqei,cn=test,ou=ae-dir\xa0/0'
            b'-\x04\x0bdisplayName1\x1e\x04\x1cFred Feuerstein (rqei/30054)0U'
            b'\x04\x06member\x04\x1auid=tmar,cn=test,ou=ae-dir\xa0/0-\x04\x0b'
            b'displayName1\x1e\x04\x1cHorst R\xc3\xbcbezahl (tmar/30082)'
        )
        self.assertIsInstance(deref.derefRes, dict)
        self.assertIsInstance(deref.derefRes['member'][0], ldap0.res.SearchResultEntry)
        self.assertEqual(
            deref.derefRes['member'][0].dn_s,
            'uid=rqei,cn=test,ou=ae-dir'
        )
        self.assertEqual(
            deref.derefRes['member'][0].entry_s,
            {'displayName': ['Fred Feuerstein (rqei/30054)']}
        )
        self.assertEqual(
            deref.derefRes['member'][1].dn_s,
            'uid=tmar,cn=test,ou=ae-dir'
        )
        self.assertEqual(
            deref.derefRes['member'][1].entry_s,
            {'displayName': ['Horst Rübezahl (tmar/30082)']}
        )
        deref = DereferenceControl(criticality=False)
        deref.decode(
            b'0\x81\x9e0U\x04\x06member\x04\x1auid=wjcg,cn=test,ou=ae-dir\xa0/0'
            b'-\x04\x0bdisplayName1\x1e\x04\x1cFred Feuerstein (wjcg/30039)0E'
            b'\x04\x06member\x04;uid=sys-test-service-1,cn=test-services-1,cn=test,ou=ae-dir'
        )
        self.assertIsInstance(deref.derefRes, dict)
        self.assertIsInstance(deref.derefRes['member'][0], ldap0.res.SearchResultEntry)
        self.assertEqual(
            deref.derefRes['member'][0].dn_s,
            'uid=wjcg,cn=test,ou=ae-dir'
        )
        self.assertEqual(
            deref.derefRes['member'][0].entry_s,
            {'displayName': ['Fred Feuerstein (wjcg/30039)']}
        )
        self.assertEqual(
            deref.derefRes['member'][1].dn_s,
            'uid=sys-test-service-1,cn=test-services-1,cn=test,ou=ae-dir'
        )
        self.assertEqual(
            deref.derefRes['member'][1].entry_s,
            {}
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.extop.passmod - Classes for Password Modify extended operation
(see RFC 3062)
"""

# Imports from pyasn1
from pyasn1.type import namedtype, univ, tag
from pyasn1.codec.der import encoder, decoder
from pyasn1_modules.rfc2251 import LDAPString

from . import ExtendedRequest, ExtendedResponse


class UserIdentity(LDAPString):
    """
    userIdentity [0] OCTET STRING OPTIONAL
    """
    tagSet = LDAPString.tagSet.tagImplicitly(
        tag.Tag(
            tag.tagClassContext,
            tag.tagFormatSimple,
            0
        )
    )

class OldPasswd(LDAPString):
    """
    oldPasswd [1] OCTET STRING OPTIONAL
    """
    tagSet = LDAPString.tagSet.tagImplicitly(
        tag.Tag(
            tag.tagClassContext,
            tag.tagFormatSimple,
            1
        )
    )

class NewPasswd(LDAPString):
    """
    newPasswd [2] OCTET STRING OPTIONAL
    """
    tagSet = LDAPString.tagSet.tagImplicitly(
        tag.Tag(
            tag.tagClassContext,
            tag.tagFormatSimple,
            2
        )
    )


class PassmodRequest(ExtendedRequest):
    """
    Password Modify extended request
    """
    __slots__ = (
        'userIdentity',
        'oldPasswd',
        'newPasswd',
    )
    requestName: str = '1.3.6.1.4.1.4203.1.11.1'

    class PasswdModifyRequestValue(univ.Sequence):
        """
        PasswdModifyRequestValue ::= SEQUENCE {
          userIdentity    [0]  OCTET STRING OPTIONAL,
          oldPasswd       [1]  OCTET STRING OPTIONAL,
          newPasswd       [2]  OCTET STRING OPTIONAL }
        """
        componentType = namedtype.NamedTypes(
            namedtype.OptionalNamedType('userIdentity', UserIdentity()),
            namedtype.OptionalNamedType('oldPasswd', OldPasswd()),
            namedtype.OptionalNamedType('newPasswd', NewPasswd()),
        )

    def __init__(self, userIdentity=None, oldPasswd=None, newPasswd=None):
        ExtendedRequest.__init__(self)
        self.userIdentity = userIdentity
        self.oldPasswd = oldPasswd
        self.newPasswd = newPasswd

    def encode(self) -> bytes:
        req_value = self.PasswdModifyRequestValue()
        req_value.setComponentByName(
            'userIdentity',
            UserIdentity(self.userIdentity.encode(self.encoding)),
        )
        if self.oldPasswd is not None:
            req_value.setComponentByName(
                'oldPasswd',
                OldPasswd(self.oldPasswd),
            )
        if self.newPasswd is not None:
            req_value.setComponentByName(
                'newPasswd',
                NewPasswd(self.newPasswd),
            )
        return encoder.encode(req_value)


class GenPasswd(LDAPString):
    """
    genPasswd       [0]     OCTET STRING OPTIONAL
    """
    tagSet = LDAPString.tagSet.tagImplicitly(
        tag.Tag(
            tag.tagClassContext,
            tag.tagFormatSimple,
            0
        )
    )
    encoding = 'utf-8'


class PasswdModifyResponseValue(univ.Sequence):
    """
    PasswdModifyResponseValue ::= SEQUENCE {
       genPasswd [0] OCTET STRING OPTIONAL }
    """
    componentType = namedtype.NamedTypes(
        namedtype.OptionalNamedType(
            'genPasswd',
            GenPasswd(),
        )
    )


class PassmodResponse(ExtendedResponse):
    """
    Password Modify extended response
    """
    __slots__ = (
        'genPasswd',
    )
    responseName = None

    def decode(self, value: bytes):
        self.genPasswd = None
        if value is None:
            return
        passmod_resp, _ = decoder.decode(
            value,
            asn1Spec=PasswdModifyResponseValue(),
        )
        if 'genPasswd' in passmod_resp:
            self.genPasswd = bytes(passmod_resp.getComponentByName('genPasswd'))



# -*- coding: ascii -*-
"""
Automatic tests for module ldap0.tmpl
"""

# from Python's standard lib
import unittest

from ldap0.tmpl import EscapeStrDict, TemplateEntry
import ldap0.dn


class TestEscapeStrDict(unittest.TestCase):
    """
    test ldap0.tmpl.EscapeStrDict
    """

    def test001_escape_dn(self):
        escaped_cn = ldap0.dn.escape_str('Foo=Bar')
        edct1 = EscapeStrDict(ldap0.dn.escape_str, cn='Foo=Bar', mail='theo@example.com')
        self.assertEqual(edct1['cn'], escaped_cn)
        self.assertEqual(edct1['mail'], 'theo@example.com')
        self.assertEqual(sorted(edct1.values()), [escaped_cn, 'theo@example.com'])
        self.assertEqual(sorted(edct1.items()), [('cn', escaped_cn), ('mail', 'theo@example.com')])
        edct2 = EscapeStrDict(ldap0.dn.escape_str, {'cn': 'Foo=Bar'})
        self.assertEqual(edct2['cn'], escaped_cn)
        edct3 = EscapeStrDict(ldap0.dn.escape_str, (('cn', 'Foo=Bar'),))
        self.assertEqual(edct3['cn'], escaped_cn)


class TestTemplateEntry(unittest.TestCase):
    """
    test ldap0.tmpl.TemplateEntry
    """

    def test001_inetorgperson_ascii(self):
        t_dn = 'cn={first_name} {last_name},ou=People,dc=example,dc=com'
        t_entry = {
            'objectClass': ['top', 'person', 'organizationalPerson', 'inetOrgPerson'],
            'cn': ['{first_name} {last_name}', '{last_name}, {first_name}'],
            'sn': ['{last_name}'],
            'mail': ['{mail_addr}'],
            'departmentNumber': ['{dept_num:08d}'],
        }
        inet_org_person = TemplateEntry(t_dn=t_dn, t_entry=t_entry)
        self.assertEqual(inet_org_person.t_dn, t_dn)
        self.assertEqual(inet_org_person.t_entry, t_entry)
        self.assertEqual(
            inet_org_person.ldap_entry({
                'first_name': 'Theo',
                'last_name': 'Tester',
                'mail_addr': 'theo.tester@example.com',
                'dept_num': 42,
            }),
            (
                'cn=Theo Tester,ou=People,dc=example,dc=com',
                {
                    'objectClass': [b'top', b'person', b'organizationalPerson', b'inetOrgPerson'],
                    'cn': [b'Theo Tester', b'Tester, Theo'],
                    'sn': [b'Tester'],
                    'mail': [b'theo.tester@example.com'],
                    'departmentNumber': [b'00000042'],
                }
            )
        )
        # missing mail_addr
        self.assertEqual(
            inet_org_person.ldap_entry({
                'first_name': 'Theo',
                'last_name': 'T\xe4ster',
            }),
            (
                'cn=Theo T\xe4ster,ou=People,dc=example,dc=com',
                {
                    'objectClass': [b'top', b'person', b'organizationalPerson', b'inetOrgPerson'],
                    'cn': [b'Theo T\xc3\xa4ster', b'T\xc3\xa4ster, Theo'],
                    'sn': [b'T\xc3\xa4ster'],
                }
            )
        )

    def test002_unicode(self):
        inet_org_person = TemplateEntry(
            t_dn = 'cn={first_name} {last_name},ou=People,dc=example,dc=com',
            t_entry = {
                'objectClass': ['top', 'person', 'organizationalPerson', 'inetOrgPerson'],
                'cn': ['{first_name} {last_name}', '{last_name}, {first_name}'],
                'sn': ['{last_name}'],
                'mail': ['{mail_addr}'],
                'description': ['T\xe4stuser'],
            },
        )
        self.assertEqual(
            inet_org_person.ldap_entry({
                'first_name': 'Theo',
                'last_name': 'Tester',
                'mail_addr': 'theo.tester@example.com',
            }),
            (
                'cn=Theo Tester,ou=People,dc=example,dc=com',
                {
                    'objectClass': [b'top', b'person', b'organizationalPerson', b'inetOrgPerson'],
                    'cn': [b'Theo Tester', b'Tester, Theo'],
                    'sn': [b'Tester'],
                    'mail': [b'theo.tester@example.com'],
                    'description': [b'T\xc3\xa4stuser'],
                }
            )
        )
        # missing mail_addr
        self.assertEqual(
            inet_org_person.ldap_entry({
                'first_name': 'Theo',
                'last_name': 'T\xe4ster',
            }),
            (
                'cn=Theo T\xe4ster,ou=People,dc=example,dc=com',
                {
                    'objectClass': [b'top', b'person', b'organizationalPerson', b'inetOrgPerson'],
                    'cn': [b'Theo T\xc3\xa4ster', b'T\xc3\xa4ster, Theo'],
                    'sn': [b'T\xc3\xa4ster'],
                    'description': [b'T\xc3\xa4stuser'],
                }
            )
        )

    def test003_unicode_first_stop(self):
        inet_org_person = TemplateEntry(
            t_dn = 'cn={first_name} {last_name},ou=People,dc=example,dc=com',
            t_entry = {
                'objectClass': ['top', 'person', 'organizationalPerson', 'inetOrgPerson'],
                'cn': ['{first_name} {last_name}', '{last_name}, {first_name}'],
                'sn': ['{last_name}'],
                'mail': ['{mail_addr}'],
                'description': ['T\xe4stuser'],
                'departmentNumber': [
                    '{no1}',
                    '{dept_num:08d}',
                ],
                'displayName': [
                    '{no1}',
                    '{first_name} {last_name} ({dept_num:08d})',
                    '{no2}',
                ],
            },
        )
        self.assertEqual(
            inet_org_person.ldap_entry(
                {
                    'first_name': 'Theo',
                    'last_name': 'Tester',
                    'mail_addr': 'theo.tester@example.com',
                    'dept_num': 42,
                },
                first_stop_attrs=('departmentNumber', 'displayName'),
            ),
            (
                'cn=Theo Tester,ou=People,dc=example,dc=com',
                {
                    'objectClass': [b'top', b'person', b'organizationalPerson', b'inetOrgPerson'],
                    'cn': [b'Theo Tester', b'Tester, Theo'],
                    'sn': [b'Tester'],
                    'displayName': [b'Theo Tester (00000042)'],
                    'mail': [b'theo.tester@example.com'],
                    'departmentNumber': [b'00000042'],
                    'description': [b'T\xc3\xa4stuser'],
                }
            )
        )

    def test004_escape_dn(self):
        self.assertEqual(
            TemplateEntry(
                t_dn = 'cn={foo},dc=example,dc=com',
                t_entry = {
                    'cn': ['{foo}'],
                }
            ).ldap_entry({'foo': 'Tester, Theo'}),
            (
                'cn=Tester\\, Theo,dc=example,dc=com',
                {'cn': [b'Tester, Theo']},
            ),
        )

    def test005_from_ldif_str(self):
        self.assertEqual(
            TemplateEntry.from_ldif_str((
                b'dn: cn={foo},dc=example,dc=com\n'
                b'cn: {foo}\n'
                b'\n'
            )).ldap_entry({'foo': 'Tester, Theo'}),
            (
                'cn=Tester\\, Theo,dc=example,dc=com',
                {'cn': [b'Tester, Theo']},
            ),
        )

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Automatic tests for ldap0
"""

from . import test_about
from . import test_base
from . import test_cache
from . import test_cext
from . import test_cidict
from . import test_controls
from . import test_dn
from . import test_filter
from . import test_functions
from . import test_modlist
from . import test_schema_parse
from . import test_ldapurl
from . import test_ldif
from . import test_ldapobject
from . import test_extensions
from . import test_ppolicy
from . import test_schema_subentry
from . import test_openldap
from . import test_extop



# -*- coding: ascii -*-
"""
ldap0.controls.pwdpolicy - classes for Password Policy controls
(see https://tools.ietf.org/html/draft-vchu-ldap-pwd-policy)
"""

from . import ResponseControl

__all__ = [
    'PasswordExpiringControl',
    'PasswordExpiredControl',
]


class PasswordExpiringControl(ResponseControl):
    """
    Indicates time in seconds when password will expire
    """
    controlType: str = '2.16.840.1.113730.3.4.5'

    def decode(self, encodedControlValue: bytes):
        self.gracePeriod = int(encodedControlValue)


class PasswordExpiredControl(ResponseControl):
    """
    Indicates that password is expired
    """
    controlType: str = '2.16.840.1.113730.3.4.4'

    def decode(self, encodedControlValue: bytes):
        self.passwordExpired = (encodedControlValue == b'0')



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.cidict
"""

# from Python's standard lib
import unittest

import ldap0.cidict


class TestCidict(unittest.TestCase):
    """
    test ldap0.cidict.CIDict
    """

    def test_cidict(self):
        data = {
            'AbCDeF':123,
        }
        cix = ldap0.cidict.CIDict(data)
        self.assertEqual(cix["ABCDEF"], 123)
        self.assertEqual(cix.get("ABCDEF", None), 123)
        self.assertEqual(cix.get("not existent", None), None)
        self.assertEqual(cix.get("not existent", 42), 42)
        cix["xYZ"] = 987
        self.assertEqual(cix["XyZ"], 987)
        self.assertEqual(cix.get("xyz", None), 987)
        with self.assertRaises(NotImplementedError):
            cix.has_key("xyz")
        self.assertTrue("xyz" in cix)
        self.assertTrue("Xyz" in cix)
        self.assertEqual(sorted(cix.keys()), ['AbCDeF', 'xYZ'])
        self.assertEqual(sorted(cix.items()), [('AbCDeF', 123), ('xYZ', 987)])
        del cix["abcdEF"]
        self.assertFalse("abcdef" in cix)
        self.assertFalse("AbCDef" in cix._keys)
        self.assertFalse("abcdef" in cix)
        self.assertFalse("abcdef" in cix)

    def test_mixed_type_fail(self):
        data = {
            'AbCDeF':123,
        }
        cix = ldap0.cidict.CIDict(data)
        with self.assertRaises(TypeError):
            cix[b'foo'] = 42
        with self.assertRaises(TypeError):
            cix = ldap0.cidict.CIDict(data, keytype=bytes)

class TestSingleValueDict(unittest.TestCase):
    """
    test ldap0.cidict.SingleValueDict
    """

    def test001_getitem(self):
        data = {
            'foo':['bar1', 'bar2', 'bar3'],
        }
        svd = ldap0.cidict.SingleValueDict(data)
        self.assertEqual(svd['foo'], 'bar1')
        self.assertEqual(list(svd.values()), ['bar1'])
        self.assertEqual(list(svd.items()), [('foo', 'bar1')])


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.dn - misc stuff for handling distinguished names (see RFC 4514)
"""

from typing import List, Union, Sequence, Tuple

import _libldap0

from .functions import _libldap0_function_call
from .typehints import StrList, EntryStr


# to be used with str.translate in escape_str()
BASIC_ESCAPE_MAP = {
    ord('\\'): '\\\\',
    ord(','): '\\,',
    ord('+'): '\\+',
    ord('"'): '\\"',
    ord('<'): '\\<',
    ord('>'): '\\>',
    ord(';'): '\\;',
    ord('='): '\\=',
    0: '\\\000',
}


def escape_str(val: str) -> str:
    """
    Escape all DN special characters found in val
    according to rules defined in RFC 4514, section 2.4
    """
    assert isinstance(val, str), TypeError('Expected str (unicode) for val, got %r' % (val,))
    if not val:
        return val
    val = val.translate(BASIC_ESCAPE_MAP)
    if val[-1:] == ' ':
        val = ''.join((val[:-1], '\\ '))
    if val[0:1] == '#' or val[0:1] == ' ':
        val = ''.join(('\\', val))
    return val


def is_dn(val: str, flags: int = 0) -> bool:
    """
    Returns True is `s' can be parsed by _libldap0.str2dn() like as a
    distinguished host_name (DN), otherwise False is returned.
    """
    assert isinstance(val, str), TypeError('Expected str (unicode) for val, got %r' % (val,))
    if not val:
        # _libldap0.str2dn() seg faults on empty bytes sequence
        # => return immediaetly here
        return True
    try:
        _libldap0_function_call(_libldap0.str2dn, val.encode('utf-8'), flags)
    except Exception:
        return False
    else:
        return True


class DNObj(tuple):
    """
    Class for handling LDAPv3 distinguished name string representations
    """
    encoding = 'utf-8'

    def __init__(self, dn_t: Sequence[Tuple[Tuple[str, str]]]):
        tuple.__init__(self)
        self = tuple(dn_t)

    @classmethod
    def from_str(cls, dn: str, flags=0, at_sanitizer=(lambda at: at)):
        """
        create a new DNObj instance from DN string representation
        """
        if not dn:
            return cls([])
        return cls([
            tuple([
                (
                    at_sanitizer(at.decode('ascii')),
                    (None if av is None else av.decode(cls.encoding))
                )
                for at, av, _ in dn_comp
            ])
            for dn_comp in _libldap0_function_call(_libldap0.str2dn, dn.encode(cls.encoding), flags)
        ])

    @classmethod
    def from_domain(cls, domain: str):
        """
        create a new DNObj instance from dot-separated DNS domain name
        """
        domain = domain.encode('idna').decode('ascii')
        return cls([(('dc', d),) for d in domain.split('.')])

    def encode(self, encoding=None, errors='strict') -> bytes:
        return self.__str__().encode(encoding or self.encoding, errors=errors)

    def __str__(self) -> str:
        return ','.join([
            '+'.join([
                '='.join((atype, escape_str(avalue or '')))
                for atype, avalue in rdn
            ])
            for rdn in self
        ])

    def __bytes__(self) -> bytes:
        return self.encode()

    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self) -> str:
        return '%s(%s)' % (self.__class__.__name__, tuple.__repr__(self))

    def __eq__(self, other) -> bool:
        res = isinstance(other, self.__class__) and len(self) == len(other)
        i = 0
        while res and i < len(self):
            if len(self[i]) != len(other[i]):
                res = False
                break
            j = 0
            other_dict = {
                at.lower(): av
                for at, av in other[i]
            }
            while res and j < len(self[i]):
                at_lower = self[i][j][0].lower()
                try:
                    other_val = other_dict[at_lower]
                except KeyError:
                    res = False
                else:
                    res = self[i][j][1] == other_dict[at_lower]
                j += 1
            i += 1
        return res

    def __add__(self, other):
        return self.__class__(tuple.__add__(self, other))

    def slice(self, ind1: Union[int, None], ind2: Union[int, None]):
        return self.__class__(self[ind1:ind2])

    def __reversed__(self):
        return self.__class__(reversed(tuple(self)))

    def rdn(self):
        """
        returns the RDN as DNObj instance
        """
        return self.__class__(self[:1])

    def rdn_attrs(self) -> EntryStr:
        """
        returns the RDN attributes as string-keyed dict
        """
        if not self:
            return {}
        return dict([(at, av) for at, av in self[0]])

    def parent(self):
        """
        returns the parent's DN as DNObj instance
        """
        return self.__class__(self[1:])

    def parents(self) -> List:
        """
        returns the parent's DN as DNObj instance
        """
        if len(self) <= 0:
            return []
        if len(self) == 1:
            return [self.__class__((()))]
        return [
            self.__class__(self[i:])
            for i in range(1, len(self))
        ]

    def match(self, dn_list: StrList):
        """
        returns best matching parent DN found in dn_list
        """
        max_match_level = 0
        max_match_dn = self.__class__((()))
        if not self or not dn_list:
            return max_match_dn
        self_len = len(self)
        for dn_obj in dn_list:
            dn_len = len(dn_obj)
            if dn_len > self_len:
                # can't match
                continue
            match_level = 0
            while (
                    match_level <= self_len
                    and (
                        self.slice(self_len-match_level-1, self_len-match_level)
                        == dn_obj.slice(dn_len-match_level-1, dn_len-match_level)
                    )
                ):
                match_level += 1
            if match_level > max_match_level:
                max_match_level = match_level
                max_match_dn = self.slice(-match_level, None)
        return max_match_dn

    def domain(self, only_dc: bool = True) -> str:
        """
        Convert dc-style DN to DNS domain name (see RFC 2247)
        """
        dns_labels = []
        for ind in range(len(self)-1, -1, -1):
            dn_comp = self[ind]
            if len(dn_comp) != 1:
                raise ValueError('multi-valued RDN not allowed for domain, was %r' % (dn_comp,))
            attr_type, attr_value = dn_comp[0]
            if attr_type.lower() != 'dc':
                if only_dc:
                    raise ValueError('Only dc allowed as RDN attribute, was %r' % (attr_type,))
                break
            dns_labels.append(attr_value.strip())
        dns_labels.reverse()
        return '.'.join(dns_labels)



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.functions
"""

# from Python's standard lib
import unittest
import datetime

import ldap0.functions
import ldap0.dn
import ldap0.filter


class TestFunction(unittest.TestCase):
    """
    test simple functions in ldap0.functions
    """

    def test_strf_secs(self):
        self.assertEqual(
            ldap0.functions.strf_secs(0),
            '19700101000000Z'
        )
        self.assertEqual(
            ldap0.functions.strf_secs(1466947067),
            '20160626131747Z'
        )

    def test_strp_secs(self):
        self.assertEqual(
            ldap0.functions.strp_secs('19700101000000Z'),
            0
        )
        self.assertEqual(
            ldap0.functions.strp_secs('20160626131747Z'),
            1466947067
        )

    def test_escape_str(self):
        self.assertEqual(
            ldap0.functions.escape_str(
                ldap0.filter.escape_str,
                '(&(objectClass=aeUser)(uid=%s))',
                'foo'
            ),
            '(&(objectClass=aeUser)(uid=foo))'
        )
        self.assertEqual(
            ldap0.functions.escape_str(
                ldap0.filter.escape_str,
                '(&(objectClass=aeUser)(uid=%s))',
                'foo)bar'
            ),
            '(&(objectClass=aeUser)(uid=foo\\29bar))'
        )
        self.assertEqual(
            ldap0.functions.escape_str(
                ldap0.dn.escape_str,
                'uid=%s',
                'foo=bar'
            ),
            'uid=foo\\=bar'
        )
        self.assertEqual(
            ldap0.functions.escape_str(
                ldap0.dn.escape_str,
                'uid=%s,cn=%s,cn=%s,dc=example,dc=com',
                'foo=bar',
                'foo+',
                '+bar',
            ),
            'uid=foo\\=bar,cn=foo\\+,cn=\\+bar,dc=example,dc=com'
        )

    def test_attr_set(self):
        self.assertEqual(
            ldap0.functions.attr_set('foo bar foobar'),
            {'foobar', 'foo', 'bar'}
        )
        self.assertEqual(
            ldap0.functions.attr_set('foo,bar,foobar'),
            {'foobar', 'foo', 'bar'}
        )
        self.assertEqual(
            ldap0.functions.attr_set('    foo, bar, foobar'),
            {'foobar', 'foo', 'bar'}
        )
        self.assertEqual(
            ldap0.functions.attr_set('    foo, bar, foo bar'),
            {'foo', 'bar'}
        )

    def test_str2datetime(self):
        self.assertEqual(
            ldap0.functions.str2datetime('19700101000000Z'),
            datetime.datetime(1970, 1, 1)
        )
        self.assertEqual(
            ldap0.functions.str2datetime('20200328123456Z'),
            datetime.datetime(2020, 3, 28, 12, 34, 56)
        )

    def test_datetime2str(self):
        self.assertEqual(
            '19700101000000Z',
            ldap0.functions.datetime2str(datetime.datetime(1970, 1, 1))
        )
        self.assertEqual(
            '20200328123456Z',
            ldap0.functions.datetime2str(datetime.datetime(2020, 3, 28, 12, 34, 56))
        )

    def test_is_expired(self):
        self.assertTrue(
            ldap0.functions.is_expired(
                datetime.datetime.utcnow()-datetime.timedelta(seconds=2),
                datetime.timedelta(seconds=1),
            )
        )
        for start_time in (
               datetime.datetime(1970, 1, 1),
               '19700101000000Z',
               b'19700101000000Z',
               0,
               0.0,
            ):
            for max_age_1, max_age_0 in (
                    (datetime.timedelta(seconds=1), datetime.timedelta(seconds=0)),
                    ('1', '0'),
                    (1, 0),
                    (1.0, 0.0),
                ):
                self.assertTrue(
                    ldap0.functions.is_expired(
                        start_time,
                        max_age_1,
                    )
                )
                self.assertFalse(
                    ldap0.functions.is_expired(
                        start_time,
                        max_age_0,
                        disable_secs=0,
                    )
                )
                self.assertTrue(
                    ldap0.functions.is_expired(
                        start_time,
                        max_age_0,
                        disable_secs=-1,
                    )
                )


class TestOptions(unittest.TestCase):
    """
    test LDAP option functions in ldap0.functions
    """

    def test_protocol_version(self):
        # default should be LDAPv3
        self.assertEqual(ldap0.functions.get_option(ldap0.OPT_PROTOCOL_VERSION), 3)
        # switch to LDAPv2
        ldap0.functions.set_option(ldap0.OPT_PROTOCOL_VERSION, 2)
        self.assertEqual(ldap0.functions.get_option(ldap0.OPT_PROTOCOL_VERSION), 2)
        # set back to LDAPv3
        ldap0.functions.set_option(ldap0.OPT_PROTOCOL_VERSION, 3)
        self.assertEqual(ldap0.functions.get_option(ldap0.OPT_PROTOCOL_VERSION), 3)

    def test_ca_cert(self):
        ldap0.functions.set_option(ldap0.OPT_X_TLS_CACERTFILE, b'/etc/ssl/ca-bundle.pem')
        self.assertEqual(ldap0.functions.get_option(ldap0.OPT_X_TLS_CACERTFILE), b'/etc/ssl/ca-bundle.pem')
        ldap0.functions.set_option(ldap0.OPT_X_TLS_CACERTDIR, b'/etc/ssl/cacerts')
        self.assertEqual(ldap0.functions.get_option(ldap0.OPT_X_TLS_CACERTDIR), b'/etc/ssl/cacerts')


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.controls.openldap - classes for OpenLDAP-specific controls
"""

from pyasn1.type import univ
from pyasn1.codec.ber import decoder

from . import ResponseControl
from .simple import ValueLessRequestControl

__all__ = [
    'SearchNoOpControl',
]


class SearchNoOpControl(ValueLessRequestControl, ResponseControl):
    """
    No-op control attached to search operations implementing sort of a
    count operation

    see https://bugs.openldap.org/show_bug.cgi?id=6598#followup4
    """
    controlType: str = '1.3.6.1.4.1.4203.666.5.18'

    class SearchNoOpControlValue(univ.Sequence):
        pass

    def decode(self, encodedControlValue: bytes):
        decodedValue, _ = decoder.decode(
            encodedControlValue,
            asn1Spec=self.SearchNoOpControlValue(),
        )
        self.resultCode = int(decodedValue[0])
        self.numSearchResults = int(decodedValue[1])
        self.numSearchContinuations = int(decodedValue[2])



# -*- coding: ascii -*-
"""
ldap0.ldapobject - wrapper classes above _libldap0.LDAPObject
"""

import sys
import time
import logging
from os import strerror
from typing import Any, Iterator, List, Optional, Tuple, Union

import _libldap0
from _libldap0 import LDAPError

from .err import \
    PasswordPolicyChangeAfterReset, \
    PasswordPolicyExpirationWarning, \
    PasswordPolicyExpiredError
from . import LIBLDAP_TLS_PACKAGE
from .controls.simple import AuthorizationIdentityResponseControl
from .err import NoUniqueEntry
from .base import encode_list
from .dn import DNObj
from .res import SearchResultEntry, LDAPResult
from .lock import LDAPLock
from .functions import _libldap0_function_call
from .schema.subentry import SCHEMA_ATTRS
from .controls.openldap import SearchNoOpControl
from .controls.ppolicy import PasswordPolicyControl
from .res import encode_request_ctrls, decode_response_ctrls
from .extop import ExtendedRequest, ExtendedResponse
from .extop.dds import RefreshRequest, RefreshResponse
from .extop.whoami import WhoAmIRequest, WhoAmIResponse
from .extop.passmod import PassmodRequest, PassmodResponse
from .extop.cancel import CancelRequest
from .sasl import SaslNoninteractiveAuth
from .modlist import add_modlist, modify_modlist
from .ldapurl import LDAPUrl
from .typehints import AttrList, EntryMixed, EntryStr, RequestControls
from .cache import Cache

if __debug__:
    # Tracing is only supported in debugging mode
    import pprint

__all__ = [
    'LDAPObject',
    'LDAPObject',
    'ReconnectLDAPObject',
]

NO_FINAL_RESULT_TYPES = {
    _libldap0.RES_SEARCH_ENTRY,
    _libldap0.RES_SEARCH_REFERENCE,
    _libldap0.RES_INTERMEDIATE,
}

#-----------------------------------------------------------------------
# the real connection classes
#-----------------------------------------------------------------------

class LDAPObject:
    """
    Wrapper class around _libldap0.LDAPObject
    """
    __slots__ = (
        '_cache',
        '_cache_ttl',
        'encoding',
        '_l',
        '_libldap0_lock',
        '_msgid_funcs',
        '_outstanding_requests',
        'res_call_args',
        'timeout',
        '_trace_level',
        'uri',
        '_whoami_dn',
    )

    def __init__(
            self,
            uri: str,
            trace_level: int = 0,
            cache_ttl: Union[int, float] = 0.0,
        ):
        self.encoding = 'utf-8'
        self.res_call_args = False
        self._trace_level = trace_level
        if isinstance(uri, str):
            ldap_url = LDAPUrl(uri)
        elif isinstance(uri, LDAPUrl):
            ldap_url = uri
        else:
            raise TypeError('Expected uri to be of type LDAPUrl or str, got %r' % (uri))
        self._libldap0_lock = LDAPLock(
            '_libldap0_lock in %r' % self,
            trace_level=self._trace_level,
        )
        self.uri = None
        conn_uri = ldap_url.connect_uri()
        self._l = _libldap0_function_call(_libldap0._initialize, conn_uri.encode('ascii'))
        self._msgid_funcs = {
            self._l.add_ext,
            self._l.simple_bind,
            self._l.compare_ext,
            self._l.delete_ext,
            self._l.extop,
            self._l.modify_ext,
            self._l.rename,
            self._l.search_ext,
            self._l.unbind_ext,
        }
        self._outstanding_requests = {}
        self.uri = conn_uri
        self._whoami_dn = None
        self.timeout = -1
        self.protocol_version = _libldap0.VERSION3
        self._cache_ttl = cache_ttl
        self._cache = Cache(ttl=cache_ttl)
        self.flush_cache()
        # Do not restart connections
        self.set_option(_libldap0.OPT_RESTART, False)
        # Switch off automatic alias dereferencing
        self.set_option(_libldap0.OPT_DEREF, _libldap0.DEREF_NEVER)
        # Switch off automatic referral chasing
        self.set_option(_libldap0.OPT_REFERRALS, False)
        # always require full TLS server validation
        self.set_option(_libldap0.OPT_X_TLS_REQUIRE_CERT, _libldap0.OPT_X_TLS_HARD)

    @property
    def protocol_version(self) -> int:
        return self.get_option(_libldap0.OPT_PROTOCOL_VERSION)

    @protocol_version.setter
    def protocol_version(self, val) -> None:
        return self.set_option(_libldap0.OPT_PROTOCOL_VERSION, val)

    @property
    def deref(self) -> int:
        return self.get_option(_libldap0.OPT_DEREF)

    @deref.setter
    def deref(self, val) -> None:
        return self.set_option(_libldap0.OPT_DEREF, val)

    @property
    def referrals(self) -> int:
        return self.get_option(_libldap0.OPT_REFERRALS)

    @referrals.setter
    def referrals(self, val) -> None:
        return self.set_option(_libldap0.OPT_REFERRALS, val)

    @property
    def timelimit(self) -> int:
        return self.get_option(_libldap0.OPT_TIMELIMIT)

    @timelimit.setter
    def timelimit(self, val) -> None:
        return self.set_option(_libldap0.OPT_TIMELIMIT, val)

    @property
    def sizelimit(self) -> int:
        return self.get_option(_libldap0.OPT_SIZELIMIT)

    @sizelimit.setter
    def sizelimit(self, val) -> None:
        return self.set_option(_libldap0.OPT_SIZELIMIT, val)

    @property
    def network_timeout(self) -> int:
        return self.get_option(_libldap0.OPT_NETWORK_TIMEOUT)

    @network_timeout.setter
    def network_timeout(self, val) -> None:
        return self.set_option(_libldap0.OPT_NETWORK_TIMEOUT, val)

    def _ldap_call(self, func, *args, **kwargs):
        """
        Wrapper method mainly for serializing calls into OpenLDAP libs
        and trace logs
        """
        with self._libldap0_lock:
            if __debug__:
                if self._trace_level >= 1:
                    logging.debug(
                        '%r %s - %s.%s(%s)',
                        self,
                        self.uri,
                        self.__class__.__name__,
                        func.__name__,
                        pprint.pformat((args, kwargs)),
                        exc_info=(self._trace_level >= 9),
                    )
            try:
                result = func(*args, **kwargs)
            except LDAPError as ldap_err:
                if (
                        hasattr(ldap_err, 'args')
                        and ldap_err.args
                        and 'info' not in ldap_err.args[0]
                        and 'errno' in ldap_err.args[0]
                    ):
                    ldap_err.args[0]['info'] = strerror(ldap_err.args[0]['errno']).encode('ascii')
                if __debug__ and self._trace_level >= 2:
                    logging.debug(
                        '-> LDAPError - %s',
                        ldap_err,
                    )
                raise ldap_err
            if __debug__ and self._trace_level >= 2 and func.__name__ != "unbind_ext":
                diag_msg_ok = self._l.get_option(_libldap0.OPT_DIAGNOSTIC_MESSAGE)
                if diag_msg_ok:
                    logging.debug(
                        '-> diagnosticMessage: %r',
                        diag_msg_ok,
                    )
                logging.debug('-> %s', pprint.pformat(result))
        if func in self._msgid_funcs:
            self._outstanding_requests[result] = (func, args, kwargs)
        return result

    def __enter__(self):
        return self

    def __exit__(self, *args):
        try:
            self.unbind_s()
        except LDAPError:
            pass

    def set_tls_options(
            self,
            cacert_filename=None,
            client_cert_filename=None,
            client_key_filename=None,
            req_cert=_libldap0.OPT_X_TLS_DEMAND,
            crl_check=_libldap0.OPT_X_TLS_CRL_NONE,
        ):
        """
        set TLS options for connection even with checking whether cert/key
        files are readable
        """
        # Set path names of TLS related files
        for tls_option, tls_pathname in (
                (_libldap0.OPT_X_TLS_CACERTFILE, cacert_filename),
                (_libldap0.OPT_X_TLS_CERTFILE, client_cert_filename),
                (_libldap0.OPT_X_TLS_KEYFILE, client_key_filename),
            ):
            try:
                if tls_pathname:
                    # Check whether option file can be read
                    with open(tls_pathname, 'rb'):
                        pass
                    self._l.set_option(tls_option, tls_pathname.encode('utf-8'))
            except ValueError as value_error:
                if sys.platform != 'darwin' and \
                   str(value_error) != 'ValueError: option error':
                    raise
        # Force server cert validation
        self._l.set_option(_libldap0.OPT_X_TLS_REQUIRE_CERT, req_cert)
        # CRL check
        if LIBLDAP_TLS_PACKAGE == 'OpenSSL':
            self._l.set_option(_libldap0.OPT_X_TLS_CRLCHECK, crl_check)
        # this has to be the last option set to let libldap reinitialize TLS context
        self._l.set_option(_libldap0.OPT_X_TLS_NEWCTX, 0)
        # end of set_tls_options()

    def fileno(self) -> int:
        return self.get_option(_libldap0.OPT_DESC)

    def abandon(
            self,
            msgid: int,
            req_ctrls: Optional[RequestControls] = None,
        ):
        if msgid not in self._outstanding_requests:
            raise _libldap0.NO_SUCH_OPERATION('Unexpected msgid value %s' % (msgid,))
        res = self._ldap_call(
            self._l.abandon_ext,
            msgid,
            encode_request_ctrls(req_ctrls),
        )
        del self._outstanding_requests[msgid]
        return res

    def cancel(
            self,
            cancelid: int,
            req_ctrls: Optional[RequestControls] = None,
        ) -> int:
        if cancelid not in self._outstanding_requests:
            raise _libldap0.NO_SUCH_OPERATION('Unexpected cancelid value %s' % (cancelid,))
        return self.extop(
            CancelRequest(cancelid),
            req_ctrls=req_ctrls,
        )

    def cancel_s(
            self,
            cancelid: int,
            req_ctrls: Optional[RequestControls] = None,
        ) -> LDAPResult:
        try:
            msgid = self.cancel(cancelid, req_ctrls=req_ctrls)
            res = self.result(msgid, _libldap0.MSG_ALL, self.timeout)
            assert res.rtype == _libldap0.RES_EXTENDED, ValueError(
                'Wrong result type %d' % (res.rtype,)
            )
            return res
        except _libldap0.CANCELLED:
            pass
        del self._outstanding_requests[cancelid]

    def add(
            self,
            dn: str,
            entry: EntryMixed,
            req_ctrls: Optional[RequestControls] = None,
        ) -> int:
        self.uncache(dn)
        entry = add_modlist(entry)
        return self._ldap_call(
            self._l.add_ext,
            dn.encode(self.encoding),
            entry,
            encode_request_ctrls(req_ctrls),
        )

    def add_s(
            self,
            dn: str,
            entry,
            req_ctrls: Optional[RequestControls] = None,
        ) -> LDAPResult:
        msgid = self.add(dn, entry, req_ctrls)
        res = self.result(msgid, _libldap0.MSG_ALL, self.timeout)
        assert res.rtype == _libldap0.RES_ADD, ValueError(
            'Wrong result type %d' % (res.rtype,)
        )
        return res

    def simple_bind(
            self,
            who: str = '',
            cred: bytes = b'',
            req_ctrls: Optional[RequestControls] = None,
        ) -> int:
        self.flush_cache()
        return self._ldap_call(
            self._l.simple_bind,
            who.encode(self.encoding),
            cred,
            encode_request_ctrls(req_ctrls),
        )

    def _handle_authzid_ctrl(self, bind_ctrls):
        authz_id_ctrls = [
            ctrl
            for ctrl in bind_ctrls
            if ctrl.controlType == AuthorizationIdentityResponseControl.controlType
        ]
        if authz_id_ctrls and len(authz_id_ctrls) == 1:
            authz_id = authz_id_ctrls[0].authzId.decode('utf-8')
            if authz_id.startswith('dn:'):
                self._whoami_dn = authz_id[3:]

    def simple_bind_s(
            self,
            who: str = '',
            cred: bytes = b'',
            req_ctrls: Optional[RequestControls] = None,
        ) -> LDAPResult:
        msgid = self.simple_bind(who, cred, req_ctrls)
        res = self.result(msgid, _libldap0.MSG_ALL, self.timeout)
        assert res.rtype == _libldap0.RES_BIND, ValueError('Wrong result type %d' % (res.rtype,))
        if res.ctrls:
            self._handle_authzid_ctrl(res.ctrls)
            # Extract the password policy response control and raise
            # appropriate exceptions if needed
            ppolicy_ctrls = [
                c
                for c in res.ctrls
                if c.controlType == PasswordPolicyControl.controlType
            ]
            if ppolicy_ctrls and len(ppolicy_ctrls) == 1:
                ppolicy_ctrl = ppolicy_ctrls[0]
                if ppolicy_ctrl.error == 2:
                    raise PasswordPolicyChangeAfterReset(
                        who=who,
                        desc='Password change is needed after reset!',
                    )
                if ppolicy_ctrl.timeBeforeExpiration is not None:
                    raise PasswordPolicyExpirationWarning(
                        who=who,
                        desc='Password will expire in %d seconds!' % (
                            ppolicy_ctrl.timeBeforeExpiration
                        ),
                        timeBeforeExpiration=ppolicy_ctrl.timeBeforeExpiration,
                    )
                if ppolicy_ctrl.graceAuthNsRemaining is not None:
                    raise PasswordPolicyExpiredError(
                        who=who,
                        desc='Password expired! %d grace logins left.' % (
                            ppolicy_ctrl.graceAuthNsRemaining
                        ),
                        graceAuthNsRemaining=ppolicy_ctrl.graceAuthNsRemaining,
                    )
        return res

    def sasl_interactive_bind_s(
            self,
            sasl_mech: str,
            auth,
            req_ctrls: Optional[RequestControls] = None,
            sasl_flags=_libldap0.SASL_QUIET
        ):
        self.flush_cache()
        return self._ldap_call(
            self._l.sasl_interactive_bind_s,
            sasl_mech.encode('ascii'),
            auth,
            encode_request_ctrls(req_ctrls),
            sasl_flags
        )

    def sasl_non_interactive_bind_s(
            self,
            sasl_mech: str,
            req_ctrls: Optional[RequestControls] = None,
            sasl_flags=_libldap0.SASL_QUIET,
            authz_id: str = ''
        ):
        if not authz_id:
            # short-cut allows really non-interactive SASL bind
            # without the flaky C call-backs
            return self.sasl_bind_s(sasl_mech, b'')
        auth = SaslNoninteractiveAuth(
            authz_id=authz_id,
            trace_level=self._trace_level,
        )
        return self.sasl_interactive_bind_s(sasl_mech, auth, req_ctrls, sasl_flags)

    def sasl_bind_s(
            self,
            sasl_mech: str,
            cred: bytes,
            req_ctrls: Optional[RequestControls] = None,
        ) -> LDAPResult:
        self.flush_cache()
        return self._ldap_call(
            self._l.sasl_bind_s,
            sasl_mech.encode('ascii'),
            cred,
            encode_request_ctrls(req_ctrls),
        )

    def compare(
            self,
            dn: str,
            attr: str,
            value: bytes,
            req_ctrls: Optional[RequestControls] = None,
        ) -> int:
        return self._ldap_call(
            self._l.compare_ext,
            dn.encode(self.encoding),
            attr.encode('ascii'),
            value,
            encode_request_ctrls(req_ctrls),
        )

    def compare_s(
            self,
            dn: str,
            attr: str,
            value: bytes,
            req_ctrls: Optional[RequestControls] = None,
        ) -> bool:
        msgid = self.compare(dn, attr, value, req_ctrls)
        try:
            res = self.result(msgid, _libldap0.MSG_ALL, self.timeout)
        except _libldap0.COMPARE_TRUE:
            return True
        except _libldap0.COMPARE_FALSE:
            return False
        raise _libldap0.PROTOCOL_ERROR(
            'Compare operation returned wrong result: %r' % (res,)
        )

    def delete(
            self,
            dn: str,
            req_ctrls: Optional[RequestControls] = None,
        ) -> int:
        self.uncache(dn)
        return self._ldap_call(
            self._l.delete_ext,
            dn.encode(self.encoding),
            encode_request_ctrls(req_ctrls),
        )

    def delete_s(
            self,
            dn: str,
            req_ctrls: Optional[RequestControls] = None,
        ) -> LDAPResult:
        msgid = self.delete(dn, req_ctrls)
        res = self.result(msgid, _libldap0.MSG_ALL, self.timeout)
        assert res.rtype == _libldap0.RES_DELETE, ValueError('Wrong result type %d' % (res.rtype,))
        return res

    def extop(
            self,
            extreq: ExtendedRequest,
            req_ctrls: Optional[RequestControls] = None,
        ) -> int:
        extop_req_value = extreq.encode()
        assert extop_req_value is None or isinstance(extop_req_value, bytes), TypeError(
            'Expected extop_req_value to be bytes or None, got %r' % (extop_req_value,)
        )
        return self._ldap_call(
            self._l.extop,
            extreq.requestName.encode('ascii'),
            extop_req_value,
            encode_request_ctrls(req_ctrls),
        )

    def extop_result(
            self,
            msgid: int = _libldap0.RES_ANY,
            extop_resp_class: Optional[Any] = None,
            timeout: Union[int, float] = -1,
        ) -> ExtendedResponse:
        _, _, _, _, respoid, respvalue = self._ldap_call(
            self._l.result,
            msgid,
            1,
            max(self.timeout, timeout),
            True,
            False,
            True,
        )
        if extop_resp_class is None:
            extop_response = ExtendedResponse(encodedResponseValue=respvalue)
        else:
            if not extop_resp_class.check_resp_name(
                    None if respoid is None else respoid.decode('ascii')
                ):
                raise _libldap0.PROTOCOL_ERROR(
                    "Wrong OID in extended response! Expected %s, got %s" % (
                        extop_resp_class.responseName,
                        respoid
                    )
                )
            extop_response = extop_resp_class(encodedResponseValue=respvalue)
        if __debug__ and self._trace_level >= 2:
            logging.debug('%s.extop_result(): %r', self.__class__.__name__, extop_response)
        return extop_response

    def extop_s(
            self,
            extreq: ExtendedRequest,
            req_ctrls: Optional[RequestControls] = None,
            extop_resp_class: Optional[Any] = None,
        ) -> ExtendedResponse:
        msgid = self.extop(extreq, req_ctrls)
        return self.extop_result(msgid, extop_resp_class=extop_resp_class)

    def modify(
            self,
            dn: str,
            modlist,
            req_ctrls: Optional[RequestControls] = None,
        ) -> int:
        self.uncache(dn)
        return self._ldap_call(
            self._l.modify_ext,
            dn.encode(self.encoding),
            modlist,
            encode_request_ctrls(req_ctrls),
        )

    def modify_s(
            self,
            dn: str,
            modlist,
            req_ctrls: Optional[RequestControls] = None,
        ) -> LDAPResult:
        msgid = self.modify(dn, modlist, req_ctrls)
        res = self.result(msgid, _libldap0.MSG_ALL, self.timeout)
        assert res.rtype == _libldap0.RES_MODIFY, ValueError('Wrong result type %d' % (res.rtype,))
        return res

    def ensure_entry(
            self,
            dn: Optional[str] = None,
            entry: Optional[dict] = None,
            old_base: Optional[str] = None,
            old_filter: Optional[str] = None,
            old_attrs: Optional[AttrList] = None,
            old_entry: Optional[EntryMixed] = None,
            req_ctrls: Optional[RequestControls] = None,
            del_ignore: bool = True,
        ) -> List[LDAPResult]:
        """
        Ensure existence or absence of a single LDAP entry
        """
        if entry is None:
            # ensure LDAP entry is absent
            if dn is not None:
                del_dn = dn
            elif old_base is not None:
                old_res = self.find_unique_entry(
                    old_base,
                    scope=_libldap0.SCOPE_SUBTREE,
                    filterstr=old_filter or '(objectClass=*)',
                    attrlist=['1.1']
                )
                del_dn = old_res.dn_s
            else:
                raise ValueError('Expected either dn or old_base to be not None!')
            res = []
            try:
                res.append(self.delete_s(del_dn, req_ctrls=req_ctrls))
            except _libldap0.NO_SUCH_OBJECT as ldap_err:
                if not del_ignore:
                    raise ldap_err
            return res
        # first try to read/search old entry
        if dn is None:
            raise ValueError('Parameter dn must not be None!')
        if old_entry is not None:
            # don't read old entry
            old_dn = dn
        elif old_filter is None:
            try:
                old_res = self.read_s(dn, attrlist=old_attrs)
            except _libldap0.NO_SUCH_OBJECT:
                old_dn = None
            else:
                if old_res is None:
                    old_dn = None
                else:
                    old_dn = dn
                    old_entry = old_res.entry_as
        else:
            # forced to search for the old entry
            if old_base is None:
                old_base = dn
            try:
                old_res = self.find_unique_entry(
                    old_base,
                    scope=_libldap0.SCOPE_SUBTREE,
                    filterstr=old_filter,
                    attrlist=old_attrs,
                )
            except NoUniqueEntry:
                old_dn = None
            else:
                old_dn = old_res.dn_s
                old_entry = old_res.entry_as
        if old_dn is None:
            # new entry has to be added
            return [self.add_s(dn, entry, req_ctrls=req_ctrls)]
         # existing entry has to be renamed and/or modified
        res = []
        if old_dn != dn:
            # rename/move existing entry
            dn_o = DNObj.from_str(dn)
            res.append(
                self.rename_s(
                    old_dn,
                    str(dn_o.rdn()),
                    newsuperior=str(dn_o.parent()),
                    delold=True,
                    req_ctrls=req_ctrls,
                )
            )
            # re-read entry because characteristic attribute might have changed
            old_entry = self.read_s(dn, attrlist=old_attrs).entry_as
        mod_list = modify_modlist(old_entry, entry)
        if mod_list:
            # finally really modify the existing entry
            res.append(self.modify_s(dn, mod_list, req_ctrls=req_ctrls))
        return res

    def passwd(
            self,
            user: str = None,
            oldpw: bytes = None,
            newpw: bytes = None,
            req_ctrls: Optional[RequestControls] = None,
        ) -> int:
        self.uncache(user)
        return self.extop(
            PassmodRequest(
                userIdentity=user,
                oldPasswd=oldpw,
                newPasswd=newpw,
            ),
            req_ctrls=req_ctrls,
        )

    def passwd_s(
            self,
            user: str = None,
            oldpw: bytes = None,
            newpw: bytes = None,
            req_ctrls: Optional[RequestControls] = None,
        ) -> Union[bytes, None]:
        msgid = self.passwd(
            user=user,
            oldpw=oldpw,
            newpw=newpw,
            req_ctrls=req_ctrls,
        )
        res = self.extop_result(msgid, extop_resp_class=PassmodResponse)
        if res.responseValue is None:
            return None
        return res.genPasswd

    def rename(
            self,
            dn: str,
            newrdn: str,
            newsuperior: Optional[str] = None,
            delold: bool = True,
            req_ctrls: Optional[RequestControls] = None,
        ) -> int:
        return self._ldap_call(
            self._l.rename,
            dn.encode(self.encoding),
            newrdn.encode(self.encoding),
            newsuperior.encode(self.encoding) if newsuperior is not None else None,
            delold,
            encode_request_ctrls(req_ctrls),
        )

    def rename_s(
            self,
            dn: str,
            newrdn: str,
            newsuperior: Optional[str] = None,
            delold: bool = True,
            req_ctrls: Optional[RequestControls] = None,
        ) -> LDAPResult:
        msgid = self.rename(dn, newrdn, newsuperior, delold, req_ctrls)
        res = self.result(msgid, _libldap0.MSG_ALL, self.timeout)
        assert res.rtype == _libldap0.RES_MODRDN, ValueError('Wrong result type %d' % (res.rtype,))
        return res

    def result(
            self,
            msgid: int,
            all_results: int = _libldap0.MSG_ALL,
            timeout: Union[int, float] = -1,
            add_intermediates: bool = False,
            resp_ctrl_classes=None,
        ) -> LDAPResult:
        if msgid != _libldap0.RES_ANY and msgid not in self._outstanding_requests:
            raise _libldap0.NO_SUCH_OPERATION('Unexpected msgid value %s' % (msgid,))
        try:
            res_type, res_data, res_msgid, res_ctrls = self._ldap_call(
                self._l.result,
                msgid,
                all_results,
                timeout,
                True,
                add_intermediates,
                False
            )
        except LDAPError as ldap_err:
            try:
                err_ctrls = ldap_err.args[0]['ctrls']
            except (IndexError, KeyError):
                ldap_err.ctrls = None
            else:
                ldap_err.ctrls = decode_response_ctrls(err_ctrls, resp_ctrl_classes)
            raise ldap_err
        if res_type is None:
            return
        assert msgid == _libldap0.RES_ANY or res_msgid == msgid, ValueError(
            'Expected LDAP result with msgid %d, got %d' % (msgid, res_msgid)
        )
        if self.res_call_args:
            call_args = self._outstanding_requests.get(res_msgid, None)
        else:
            call_args = None
        if res_type not in NO_FINAL_RESULT_TYPES:
            del self._outstanding_requests[res_msgid]
        return LDAPResult(
            res_type,
            [
                (dn, entry, decode_response_ctrls(ctrls, resp_ctrl_classes))
                for dn, entry, ctrls in res_data
            ],
            res_msgid,
            decode_response_ctrls(res_ctrls, resp_ctrl_classes),
            call_args=call_args,
        )

    def results(
            self,
            msgid: int,
            timeout: Union[int, float] = -1,
            add_intermediates: bool = False,
            resp_ctrl_classes=None,
        ) -> Iterator[LDAPResult]:
        """
        Generator method which returns an iterator for processing all LDAP
        operation results of the given msgid like retrieved with
        LDAPObject.result()
        """
        if msgid != _libldap0.RES_ANY and msgid not in self._outstanding_requests:
            raise _libldap0.NO_SUCH_OPERATION('Unexpected msgid value %s' % (msgid,))
        start_time = time.time()
        if timeout >= 0:
            end_time = start_time + timeout
        while timeout < 0 or time.time() <= end_time:
            try:
                res = self.result(
                    msgid,
                    all_results=_libldap0.MSG_ONE,
                    timeout=min(timeout, 0.5),
                    add_intermediates=add_intermediates,
                    resp_ctrl_classes=resp_ctrl_classes,
                )
            except _libldap0.TIMEOUT:
                if timeout >= 0 and time.time() > end_time:
                    raise _libldap0.TIMEOUT('Timeout of %0.4f secs reached.' % (timeout,))
                continue
            yield res
            if res.rtype not in NO_FINAL_RESULT_TYPES:
                break
        # end of results()

    def search(
            self,
            base: str,
            scope: int,
            filterstr: str = '(objectClass=*)',
            attrlist: Optional[AttrList] = None,
            attrsonly: bool = False,
            req_ctrls: Optional[RequestControls] = None,
            timeout: Union[int, float] = -1,
            sizelimit=0
        ) -> int:
        assert isinstance(base, str), TypeError(
            'Expected str for base, got %r' % (base,)
        )
        assert isinstance(filterstr, str), TypeError(
            'Expected str for filterstr, got %r' % (filterstr,)
        )
        if attrlist is not None:
            attrlist = encode_list(attrlist, encoding='ascii')
        else:
            attrlist = [b'*']
        return self._ldap_call(
            self._l.search_ext,
            base.encode(self.encoding),
            scope,
            filterstr.encode(self.encoding),
            attrlist,
            attrsonly,
            encode_request_ctrls(req_ctrls),
            timeout,
            sizelimit,
        )

    def cache_hit_ratio(self) -> float:
        """
        Returns percentage of cache hit ratio
        """
        return self._cache.hit_ratio

    def flush_cache(self) -> None:
        """
        reset/flush the internal cache
        """
        try:
            cache = self._cache
        except AttributeError:
            pass
        else:
            cache.flush()
        self._whoami_dn = None

    def uncache(self, entry_dn) -> None:
        """
        remove all cached entries related to dn from internal cache
        """
        entry_dn = entry_dn.lower()
        remove_keys = set()
        # first find the items to be removed from cache
        for cache_key_args in list(self._cache):
            if cache_key_args[0].lower() == entry_dn:
                remove_keys.add(cache_key_args)
                continue
            try:
                cached_results = self._cache[cache_key_args]
            except KeyError:
                # cached items expired in between
                pass
            else:
                for cached_res in cached_results:
                    if (
                            isinstance(cached_res, SearchResultEntry)
                            and cached_res.dn_s.lower() == entry_dn
                        ):
                        remove_keys.add(cache_key_args)
        # finally remove items from cache dict
        for cache_key_args in remove_keys:
            try:
                del self._cache[cache_key_args]
            except KeyError:
                pass
        # end of uncache()

    def search_s(
            self,
            base: str,
            scope: int,
            filterstr: str = '(objectClass=*)',
            attrlist: Optional[AttrList] = None,
            attrsonly: bool = False,
            req_ctrls: Optional[RequestControls] = None,
            timeout: Union[int, float] = -1,
            sizelimit=0,
            cache_ttl: Optional[Union[int, float]] = None,
        ) -> List[LDAPResult]:
        if cache_ttl is None:
            cache_ttl = self._cache._ttl
        if cache_ttl > 0:
            cache_key_args = (
                base,
                scope,
                filterstr,
                tuple(attrlist or []),
                attrsonly,
                tuple([
                    c.encode()
                    for c in req_ctrls or []
                ]),
                timeout,
                sizelimit,
            )
            # first look into cache for non-expired search results
            try:
                res = self._cache[cache_key_args]
            except KeyError:
                pass
            else:
                return res
        # no cached result
        msgid = self.search(
            base,
            scope,
            filterstr,
            attrlist,
            attrsonly,
            req_ctrls,
            timeout,
            sizelimit
        )
        res = []
        for ldap_res in self.results(msgid, timeout=timeout):
            res.extend(ldap_res.rdata)
        if cache_ttl > 0:
            # store result in cache if caching is enabled at all
            self._cache.cache(cache_key_args, res, cache_ttl)
        return res
        # search_s()

    def noop_search(
            self,
            base: str,
            scope: int = _libldap0.SCOPE_SUBTREE,
            filterstr: str = '(objectClass=*)',
            req_ctrls: Optional[RequestControls] = None,
            timeout: Union[int, float] = -1,
            sizelimit=0,
        ) -> Tuple[Optional[int], Optional[int]]:
        req_ctrls = req_ctrls or []
        req_ctrls.append(SearchNoOpControl(criticality=True))
        try:
            msg_id = self.search(
                base,
                scope,
                filterstr=filterstr,
                attrlist=['1.1'],
                req_ctrls=req_ctrls,
                timeout=timeout,
                sizelimit=sizelimit,
            )
            noop_srch_res = list(self.results(
                msg_id,
                timeout=timeout,
            ))[0]
        except (
                _libldap0.TIMEOUT,
                _libldap0.TIMELIMIT_EXCEEDED,
                _libldap0.SIZELIMIT_EXCEEDED,
                _libldap0.ADMINLIMIT_EXCEEDED
            ) as err:
            self.abandon(msg_id)
            raise err
        else:
            noop_srch_ctrl = [
                c
                for c in noop_srch_res.ctrls
                if c.controlType == SearchNoOpControl.controlType
            ]
            if noop_srch_ctrl:
                return noop_srch_ctrl[0].numSearchResults, noop_srch_ctrl[0].numSearchContinuations
            return (None, None)

    def start_tls_s(self):
        return self._ldap_call(self._l.start_tls_s)

    def unbind(self, req_ctrls: Optional[RequestControls] = None) -> int:
        res = self._ldap_call(
            self._l.unbind_ext,
            encode_request_ctrls(req_ctrls),
        )
        try:
            del self._l
        except AttributeError:
            pass
        return res

    def unbind_s(self, req_ctrls: Optional[RequestControls] = None) -> LDAPResult:
        msgid = self.unbind(req_ctrls)
        if msgid is None:
            return None
        res = self.result(msgid, _libldap0.MSG_ALL, self.timeout)
        return res

    def whoami_s(self, req_ctrls: Optional[RequestControls] = None) -> str:
        wai = self.extop_s(
            WhoAmIRequest(),
            extop_resp_class=WhoAmIResponse,
            req_ctrls=req_ctrls,
        )
        if wai.responseValue and wai.responseValue.lower().startswith('dn:'):
            # cache this for later calls to .whoami_s()
            self._whoami_dn = wai.responseValue[3:]
        return wai.responseValue

    def get_option(self, option: int) -> Union[int, float, bytes]:
        result = self._ldap_call(self._l.get_option, option)
        if option == _libldap0.OPT_SERVER_CONTROLS:
            result = decode_response_ctrls(result)
        return result

    def set_option(self, option: int, invalue: Union[int, float, bytes, RequestControls]):
        if option == _libldap0.OPT_SERVER_CONTROLS:
            invalue = encode_request_ctrls(invalue)
        return self._ldap_call(self._l.set_option, option, invalue)

    def search_subschemasubentry_s(self, dn='', rootdse=True):
        res = None
        try:
            ldap_res = self.read_s(dn, attrlist=['subschemaSubentry'])
        except (
                _libldap0.NO_SUCH_OBJECT,
                _libldap0.INSUFFICIENT_ACCESS,
                _libldap0.NO_SUCH_ATTRIBUTE,
                _libldap0.UNDEFINED_TYPE,
                _libldap0.REFERRAL,
            ):
            pass
        else:
            if ldap_res is not None and 'subschemaSubentry' in ldap_res.entry_s:
                res = ldap_res.entry_s['subschemaSubentry'][0]
        if res is None and dn and rootdse:
            # fall back reading attribute subschemaSubentry from rootDSE
            res = self.search_subschemasubentry_s(dn='')
        return res # end of search_subschemasubentry_s()

    def read_s(
            self,
            dn: str,
            filterstr: str = '(objectClass=*)',
            attrlist: Optional[AttrList] = None,
            req_ctrls: Optional[RequestControls] = None,
            timeout: Union[int, float] = -1,
            cache_ttl: Optional[Union[int, float]] = None,
        ) -> SearchResultEntry:
        ldap_res = self.search_s(
            dn,
            _libldap0.SCOPE_BASE,
            filterstr,
            attrlist=attrlist,
            req_ctrls=req_ctrls,
            timeout=timeout,
            cache_ttl=cache_ttl,
        )
        if not ldap_res:
            return None
        return ldap_res[0]

    def read_subschemasubentry_s(
            self,
            subschemasubentry_dn: str,
            attrs: Optional[AttrList] = None,
        ) -> EntryStr:
        try:
            subschemasubentry = self.read_s(
                subschemasubentry_dn,
                filterstr='(objectClass=subschema)',
                attrlist=attrs or SCHEMA_ATTRS
            )
        except _libldap0.NO_SUCH_OBJECT:
            return None
        if subschemasubentry is None:
            return None
        return subschemasubentry.entry_s

    def refresh(
            self,
            dn: str,
            ttl: Optional[Union[str, float]] = None,
        ) -> int:
        self.uncache(dn)
        return self.extop(RefreshRequest(entryName=dn, requestTtl=ttl))

    def refresh_s(
            self,
            dn: str,
            ttl: Optional[Union[str, float]] = None,
        ) -> Union[str, float]:
        self.uncache(dn)
        msgid = self.refresh(dn, ttl=ttl)
        res = self.extop_result(msgid, extop_resp_class=RefreshResponse)
        return res.responseTtl

    def find_unique_entry(
            self,
            base: str,
            scope: int = _libldap0.SCOPE_SUBTREE,
            filterstr: str = '(objectClass=*)',
            attrlist: Optional[AttrList] = None,
            attrsonly: bool = False,
            req_ctrls: Optional[RequestControls] = None,
            timeout: Union[int, float] = -1,
            cache_ttl: Optional[Union[int, float]] = None,
        ) -> LDAPResult:
        try:
            ldap_res = list(self.search_s(
                base,
                scope,
                filterstr,
                attrlist=attrlist,
                attrsonly=attrsonly,
                req_ctrls=req_ctrls,
                timeout=timeout,
                sizelimit=2,
                cache_ttl=cache_ttl,
            ))
        except _libldap0.SIZELIMIT_EXCEEDED:
            raise NoUniqueEntry('Size limit exceeded for %r' % (filterstr,))
        # strip search continuations
        ldap_res = [res for res in ldap_res if isinstance(res, SearchResultEntry)]
        # now check again for length
        if not ldap_res:
            raise NoUniqueEntry('No search result for %r' % (filterstr,))
        if len(ldap_res) > 1:
            raise NoUniqueEntry(
                'Non-unique search results (%d) for %r' % (len(ldap_res), filterstr)
            )
        return ldap_res[0]

    def read_rootdse_s(
            self,
            filterstr: str = '(objectClass=*)',
            attrlist: Optional[AttrList] = None,
        ) -> LDAPResult:
        """
        convenience wrapper around read_s() for reading rootDSE
        """
        ldap_rootdse = self.read_s(
            '',
            filterstr=filterstr,
            attrlist=attrlist or ['*', '+'],
        )
        return ldap_rootdse  # read_rootdse_s()

    def get_naming_contexts(self) -> List[str]:
        return self.read_rootdse_s(
            attrlist=['namingContexts']
        ).entry_s.get('namingContexts', [])

    def get_whoami_dn(self) -> str:
        """
        Return the result of Who Am I extended operation as plain DN
        """
        if self._whoami_dn is None:
            wai = self.whoami_s()
            if wai == '':
                self._whoami_dn = wai
            elif wai.lower().startswith('dn:'):
                self._whoami_dn = wai[3:]
            else:
                self._whoami_dn = None
                raise ValueError('Expected dn: form of Who Am I? result, got %r' % (wai,))
        return self._whoami_dn


class ReconnectLDAPObject(LDAPObject):
    """
    In case of server failure (_libldap0.SERVER_DOWN) the implementations
    of all synchronous operation methods (search_s() etc.) are doing
    an automatic reconnect and rebind and will retry the very same
    operation.

    This is very handy for broken LDAP server implementations
    (e.g. in Lotus Domino) which drop connections very often making
    it impossible to have a long-lasting control flow in the
    application.
    """
    __slots__ = (
        '_last_bind',
        '_options',
        '_reconnect_lock',
        '_reconnects_done',
        '_retry_delay',
        '_retry_max',
        '_start_tls',
    )
    __transient_attrs__ = {
        '_cache',
        '_l',
        '_libldap0_lock',
        '_reconnect_lock',
        '_last_bind',
        '_outstanding_requests',
        '_msgid_funcs',
        '_whoami_dn',
    }

    def __init__(
            self,
            uri: str,
            trace_level: int = 0,
            cache_ttl: Union[int, float] = 0.0,
            retry_max: int = 1,
            retry_delay: Union[int, float] = 60.0,
        ) -> None:
        """
        Parameters like LDAPObject.__init__() with these
        additional arguments:

        retry_max
            Maximum count of reconnect trials
        retry_delay
            Time span to wait between two reconnect trials
        """
        self._options = []
        self._last_bind = None
        LDAPObject.__init__(
            self,
            uri,
            trace_level=trace_level,
            cache_ttl=cache_ttl,
        )
        self._reconnect_lock = LDAPLock(
            '_reconnect_lock in %r' % self,
            trace_level=self._trace_level,
        )
        self._retry_max = retry_max
        self._retry_delay = retry_delay
        self._start_tls = 0
        self._reconnects_done = 0

    def __getstate__(self) -> dict:
        """
        return data representation for pickled object
        """
        state = {}
        for key in self.__slots__ + LDAPObject.__slots__:
            if key not in self.__transient_attrs__:
                state[key] = getattr(self, key)
        state['_last_bind'] = (
            self._last_bind[0].__name__,
            self._last_bind[1],
            self._last_bind[2],
        )
        return state

    def __setstate__(self, data) -> None:
        """
        set up the object from pickled data
        """
        for key, val in data.items():
            setattr(self, key, val)
        self._whoami_dn = None
        self._cache = Cache(self._cache_ttl)
        self._last_bind = (
            getattr(LDAPObject, self._last_bind[0]),
            self._last_bind[1],
            self._last_bind[2],
        )
        self._libldap0_lock = LDAPLock(
            '_libldap0_lock in %r' % self,
            trace_level=self._trace_level,
        )
        self._reconnect_lock = LDAPLock(
            '_reconnect_lock in %r' % self,
            trace_level=self._trace_level,
        )
        self.reconnect(self.uri)

    def _store_last_bind(self, method, *args, **kwargs) -> None:
        self._last_bind = (method, args, kwargs)

    def _apply_last_bind(self) -> None:
        if self._last_bind is not None:
            func, args, kwargs = self._last_bind
            func(self, *args, **kwargs)
        else:
            # Send explicit anon simple bind request to provoke
            # _libldap0.SERVER_DOWN in method reconnect()
            LDAPObject.simple_bind_s(self, '', b'')

    def _restore_options(self) -> None:
        """
        Restore all recorded options
        """
        for key, val in self._options:
            LDAPObject.set_option(self, key, val)

    def passwd_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.passwd_s, *args, **kwargs)

    def reconnect(
            self,
            uri: str,
            retry_max: int = 1,
            retry_delay: Union[int, float] = 60.0,
            reset_last_bind: bool = False,
        ) -> None:
        """
        Drop and clean up old connection completely and reconnect
        """
        self._reconnect_lock.acquire()
        if reset_last_bind:
            self._last_bind = None
        try:
            reconnect_counter = retry_max
            while reconnect_counter:
                counter_text = '%d. (of %d)' % (retry_max-reconnect_counter+1, retry_max)
                if __debug__ and self._trace_level >= 1:
                    logging.debug(
                        'Trying %s reconnect to %s...\n',
                        counter_text, uri,
                    )
                try:
                    # Do the connect
                    self._l = _libldap0_function_call(
                        _libldap0._initialize,
                        uri.encode(self.encoding),
                    )
                    self._msgid_funcs = {
                        self._l.add_ext,
                        self._l.simple_bind,
                        self._l.compare_ext,
                        self._l.delete_ext,
                        self._l.extop,
                        self._l.modify_ext,
                        self._l.rename,
                        self._l.search_ext,
                        self._l.unbind_ext,
                    }
                    self._outstanding_requests = {}
                    self._restore_options()
                    # StartTLS extended operation in case this was called before
                    if self._start_tls:
                        LDAPObject.start_tls_s(self)
                    # Repeat last simple or SASL bind
                    self._apply_last_bind()
                except (_libldap0.SERVER_DOWN, _libldap0.TIMEOUT) as ldap_error:
                    if __debug__ and self._trace_level >= 1:
                        logging.debug(
                            '%s reconnect to %s failed\n',
                            counter_text, uri
                        )
                    reconnect_counter -= 1
                    if not reconnect_counter:
                        raise ldap_error
                    if __debug__ and self._trace_level >= 1:
                        logging.debug('=> delay %s...\n', retry_delay)
                    time.sleep(retry_delay)
                    LDAPObject.unbind_s(self)
                else:
                    if __debug__ and self._trace_level >= 1:
                        logging.debug(
                            '%s reconnect to %s successful => repeat last operation\n',
                            counter_text,
                            uri,
                        )
                    self._reconnects_done += 1
                    break
        finally:
            self._reconnect_lock.release()
        # end of reconnect()

    def _apply_method_s(self, func, *args, **kwargs):
        if not hasattr(self, '_l'):
            self.reconnect(self.uri, retry_max=self._retry_max, retry_delay=self._retry_delay)
        try:
            return func(self, *args, **kwargs)
        except _libldap0.SERVER_DOWN:
            if self._retry_max <= 0:
                raise
            LDAPObject.unbind_s(self)
            # Try to reconnect
            self.reconnect(self.uri, retry_max=self._retry_max, retry_delay=self._retry_delay)
            # Re-try last operation
            return func(self, *args, **kwargs)

    def set_option(self, option: int, invalue: Union[int, float, bytes, RequestControls]):
        res = LDAPObject.set_option(self, option, invalue)
        self._options.append((option, invalue))
        return res

    def simple_bind_s(self, *args, **kwargs):
        res = self._apply_method_s(LDAPObject.simple_bind_s, *args, **kwargs)
        self._store_last_bind(LDAPObject.simple_bind_s, *args, **kwargs)
        return res

    def start_tls_s(self, *args, **kwargs):
        res = self._apply_method_s(LDAPObject.start_tls_s, *args, **kwargs)
        self._start_tls = 1
        return res

    def sasl_interactive_bind_s(self, *args, **kwargs):
        """
        sasl_interactive_bind_s(who, auth) -> None
        """
        res = self._apply_method_s(LDAPObject.sasl_interactive_bind_s, *args, **kwargs)
        self._store_last_bind(LDAPObject.sasl_interactive_bind_s, *args, **kwargs)
        return res

    def sasl_bind_s(self, *args, **kwargs):
        res = self._apply_method_s(LDAPObject.sasl_bind_s, *args, **kwargs)
        self._store_last_bind(LDAPObject.sasl_bind_s, *args, **kwargs)
        return res

    def add_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.add_s, *args, **kwargs)

    def cancel_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.cancel_s, *args, **kwargs)

    def compare_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.compare_s, *args, **kwargs)

    def delete_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.delete_s, *args, **kwargs)

    def extop_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.extop_s, *args, **kwargs)

    def modify_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.modify_s, *args, **kwargs)

    def rename_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.rename_s, *args, **kwargs)

    def search_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.search_s, *args, **kwargs)

    def whoami_s(self, *args, **kwargs):
        return self._apply_method_s(LDAPObject.whoami_s, *args, **kwargs)



# -*- coding: ascii -*-
"""
ldap0.schema.parse - Low-level parsing functions for schema element strings
"""

import re

TOKENS_FINDALL = re.compile(
    # opening parenthesis
    r"(\()"
    # or
    r"|"
    # closing parenthesis
    r"(\))"
    # or
    r"|"
    # string of length >= 1 without '$() or whitespace
    r"([^'$()\s]+)"
    # or
    r"|"
    # any string or empty string surrounded by unescaped single quotes
    # except if right quote is succeeded by alphanumeric char
    r"('(?:[^'\\]|\\.)*'(?!\w))"
    # or
    r"|"
    # residue, all non-whitespace strings
    r"([^\s]+?)",
).findall


def split_tokens(sch_str):
    """
    Returns list of syntax elements with quotes and spaces stripped.
    """
    parts = []
    parens = 0
    for opar, cpar, unquoted, quoted, residue in TOKENS_FINDALL(sch_str):
        if unquoted:
            parts.append(unquoted)
        elif quoted:
            parts.append(quoted[1:-1])
        elif opar:
            parens += 1
            parts.append(opar)
        elif cpar:
            parens -= 1
            parts.append(cpar)
        elif residue == '$':
            if not parens:
                raise ValueError("'$' outside parenthesis in %r" % (sch_str))
        else:
            raise ValueError(residue, sch_str)
    if parens:
        raise ValueError("Unbalanced parenthesis in %r" % (sch_str))
    return parts

def extract_tokens(tkl, known_tokens):
    """
    Returns dictionary of known tokens with all values
    """
    assert tkl[0].strip() == "(" and tkl[-1].strip() == ")", ValueError(tkl)
    result = dict(known_tokens)
    i = 0
    l_len = len(tkl)
    while i < l_len:
        if tkl[i] in result:
            token = tkl[i]
            i += 1 # Consume token
            if i < l_len:
                if tkl[i] in result:
                    # non-valued
                    result[token] = (())
                elif tkl[i] == "(":
                    # multi-valued
                    i += 1 # Consume left parentheses
                    start = i
                    while i < l_len and tkl[i] != ")":
                        i += 1
                    result[token] = tuple([
                        v for v in tkl[start:i] if v != '$'
                    ])
                    i += 1 # Consume right parentheses
                else:
                    # single-valued
                    result[token] = (tkl[i],)
                    i += 1 # Consume single value
        else:
            i += 1 # Consume unrecognized item
    return result



# -*- coding: ascii -*-
"""
ldap0.controls.sss - classes for Server Side Sorting
(see RFC 2891)
"""

from pyasn1.type import univ, namedtype, tag, namedval, constraint
from pyasn1.codec.ber import encoder, decoder

from . import RequestControl, ResponseControl


__all__ = [
    'SSSRequestControl',
    'SSSResponseControl',
]


class SortKeyType(univ.Sequence):
    """
    SEQUENCE {
       attributeType   AttributeDescription,
       orderingRule    [0] MatchingRuleId OPTIONAL,
       reverseOrder    [1] BOOLEAN DEFAULT FALSE }
    """
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('attributeType', univ.OctetString()),
        namedtype.OptionalNamedType(
            'orderingRule',
            univ.OctetString().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
            )
        ),
        namedtype.DefaultedNamedType(
            'reverseOrder',
            univ.Boolean(False).subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
            )
        )
    )


class SortKeyListType(univ.SequenceOf):
    """
    SortKeyList ::= SEQUENCE OF SEQUENCE {
                     attributeType   AttributeDescription,
                     orderingRule    [0] MatchingRuleId OPTIONAL,
                     reverseOrder    [1] BOOLEAN DEFAULT FALSE }
    """
    componentType = SortKeyType()


class SSSRequestControl(RequestControl):
    """
    Order result server side

    >>> s = SSSRequestControl(ordering_rules=['-cn'])
    """
    controlType: str = '1.2.840.113556.1.4.473'

    def __init__(
            self,
            criticality: bool = False,
            ordering_rules=None,
        ):
        RequestControl.__init__(self, self.controlType, criticality)
        self.ordering_rules = ordering_rules
        for rule in ordering_rules:
            rule = rule.split(':')
            assert len(rule) < 3, 'syntax for ordering rule: [-]<attribute-type>[:ordering-rule]'

    def encode(self) -> bytes:
        p = SortKeyListType()
        for i, rule in enumerate(self.ordering_rules):
            q = SortKeyType()
            reverse_order = rule.startswith('-')
            if reverse_order:
                rule = rule[1:]
            try:
                attribute_type, ordering_rule = rule.split(':')
            except ValueError:
                attribute_type, ordering_rule = rule, None
            q.setComponentByName('attributeType', attribute_type.encode(self.encoding))
            if ordering_rule:
                q.setComponentByName('orderingRule', ordering_rule.encode(self.encoding))
            if reverse_order:
                q.setComponentByName('reverseOrder', 1)
            p.setComponentByPosition(i, q)
        return encoder.encode(p)


class SortResultType(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType(
            'sortResult',
            univ.Enumerated().subtype(
                namedValues=namedval.NamedValues(
                    ('success', 0),
                    ('operationsError', 1),
                    ('timeLimitExceeded', 3),
                    ('strongAuthRequired', 8),
                    ('adminLimitExceeded', 11),
                    ('noSuchAttribute', 16),
                    ('inappropriateMatching', 18),
                    ('insufficientAccessRights', 50),
                    ('busy', 51),
                    ('unwillingToPerform', 53),
                    ('other', 80)
                ),
                subtypeSpec=univ.Enumerated.subtypeSpec + \
                    constraint.SingleValueConstraint(0, 1, 3, 8, 11, 16, 18, 50, 51, 53, 80)
            )
        ),
        namedtype.OptionalNamedType(
            'attributeType',
            univ.OctetString().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
            )
        )
    )


class SSSResponseControl(ResponseControl):
    controlType: str = '1.2.840.113556.1.4.474'

    def __init__(self, criticality=False):
        ResponseControl.__init__(self, self.controlType, criticality)

    def decode(self, encodedControlValue: bytes):
        p, rest = decoder.decode(encodedControlValue, asn1Spec=SortResultType())
        assert not rest, 'all data could not be decoded'
        sort_result = p.getComponentByName('sortResult')
        self.sortResult = int(sort_result)
        attribute_type = p.getComponentByName('attributeType')
        if attribute_type.hasValue():
            self.attributeType = attribute_type
        else:
            self.attributeType = None



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.msad
"""

# from Python's standard lib
import unittest
import uuid
from binascii import unhexlify

from ldap0.msad import sid2sddl, sddl2sid


class Test001Functions(unittest.TestCase):

    test_vectors = (
        (
            b'\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xbaZ\x17^U\xdf\x83a)XU\x0eP\x04\x00\x00',
            'S-1-5-21-1578588858-1636032341-240474153-1104'
        ),
        (
            b'\x01\x04\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xbaZ\x17^U\xdf\x83a)XU\x0e',
            'S-1-5-21-1578588858-1636032341-240474153'
        ),
    )

    def test001_sid2sddl(self):
        for sid_b, sddl_s in self.test_vectors:
            self.assertEqual(sid2sddl(sid_b), sddl_s)

    def test002_sddl2sid(self):
        for sid_b, sddl_s in self.test_vectors:
            self.assertEqual(sddl2sid(sddl_s), sid_b)

    def test003_inverse(self):
        for sid_b, sddl_s in self.test_vectors:
            self.assertEqual(sid2sddl(sddl2sid(sddl_s)), sddl_s)
            self.assertEqual(sddl2sid(sid2sddl(sid_b)), sid_b)


class Test002ObjectGUID(unittest.TestCase):
    """
    test whether uuid.UUID(bytes_le=...) generates the same UUID string
    representations like displayed by ADSI-Edit
    """

    test_vectors = (
        (
            unhexlify('17:98:44:30:D2:D6:FA:4D:98:34:C6:E4:11:A2:83:FC'.replace(':', '')),
            '30449817-d6d2-4dfa-9834-c6e411a283fc',
        ),
        (
            unhexlify('F8:5F:48:1B:41:A8:77:47:85:3C:9A:63:58:71:1B:6A'.replace(':', '')),
            '1b485ff8-a841-4777-853c-9a6358711b6a',
        ),
    )

    def test001(self):
        for id_b, uuid_s in self.test_vectors:
            self.assertEqual(str(uuid.UUID(bytes_le=id_b)), uuid_s)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.syncrepl
"""

# from Python's standard lib
import unittest

from ldap0 import SCOPE_SUBTREE, SCOPE_ONELEVEL
from ldap0.ldapurl import LDAPUrl
from ldap0.openldap import SyncReplDesc, ldapsearch_cmd


class TestSyncReplDesc(unittest.TestCase):
    """
    test ldap0.openldap.SyncReplDesc
    """

    def test001_syncrepldesc_class(self):
        srd_str_list = [
            'rid=042',
            'provider=ldaps://ae-dir-deb-p2.snet1.example.com',
            'bindmethod=Sasl',
            'timeout=5',
            'network-timeout=5',
            'saslmech=external',
            'keepalive=240:10:30',
            'starttls=no',
            'tls_cert="/opt/ae-dir/etc/tls/ae-dir-deb-p1.snet1.example.com.crt"',
            'tls_key="/opt/ae-dir/etc/tls/ae-dir-deb-p1.snet1.example.com.key"',
            'tls_cacert="/opt/ae-dir/etc/tls/my-ae-dir-testca-2017-06.pem"',
            'tls_reqcert=demand',
            'tls_cipher_suite=ECDHE-RSA-AES256-GCM-SHA384:DHE-RSA-AES256-GCM-SHA384:ECDH-RSA-AES256-GCM-SHA384:!ADH',
            'tls_protocol_min=3.3',
            'tls_crlcheck=none',
            'filter="(objectClass=*)"',
            'searchbase="dc=ae-dir,dc=example,dc=org"',
            'scope=sub',
            'schemachecking=on',
            'type=refreshAndPersist',
            'retry="30 +"'
        ]
        srd = SyncReplDesc(' '.join(srd_str_list))
        self.assertEqual(srd.rid, '042')
        self.assertEqual(srd.provider, 'ldaps://ae-dir-deb-p2.snet1.example.com')
        self.assertEqual(srd.bindmethod, 'sasl')
        self.assertEqual(srd.timeout, 5)
        self.assertEqual(srd.network_timeout, 5)
        self.assertEqual(srd.saslmech, 'EXTERNAL')
        self.assertEqual(srd.keepalive, '240:10:30')
        self.assertEqual(srd.starttls, 'no')
        self.assertEqual(srd.tls_cert, '/opt/ae-dir/etc/tls/ae-dir-deb-p1.snet1.example.com.crt')
        self.assertEqual(srd.tls_key, '/opt/ae-dir/etc/tls/ae-dir-deb-p1.snet1.example.com.key')
        self.assertEqual(srd.tls_cacert, '/opt/ae-dir/etc/tls/my-ae-dir-testca-2017-06.pem')
        self.assertEqual(srd.tls_reqcert, 'demand')
        self.assertEqual(srd.tls_cipher_suite, 'ECDHE-RSA-AES256-GCM-SHA384:DHE-RSA-AES256-GCM-SHA384:ECDH-RSA-AES256-GCM-SHA384:!ADH')
        self.assertEqual(srd.tls_protocol_min, '3.3')
        self.assertEqual(srd.tls_crlcheck, 'none')
        self.assertEqual(srd.filter, '(objectClass=*)')
        self.assertEqual(srd.searchbase, 'dc=ae-dir,dc=example,dc=org')
        self.assertEqual(srd.scope, SCOPE_SUBTREE)
        self.assertEqual(srd.schemachecking, 'on')
        self.assertEqual(srd.type, 'refreshAndPersist')
        self.assertEqual(srd.retry, '30 +')
        self.assertEqual(
            str(srd.ldap_url()),
            'ldaps://ae-dir-deb-p2.snet1.example.com/dc%3Dae-dir%2Cdc%3Dexample%2Cdc%3Dorg?*,+?sub?%28objectClass%3D%2A%29'
        )

    def test002_syncrepldesc_class(self):
        srd_str_list = [
            '',
            '',
            'rid=23',
            'provider=ldap://ldap-23.example.com',
            'bindmethod=simple',
            'timeout="5"',
            'network-timeout=5',
            'saslmech=EXTERNAL',
            'keepalive=240:10:30',
            'starttls=critical',
            'tls_cacert="/etc/ssl/ca-certs.pem"',
            'tls_reqcert=try',
            'tls_crlcheck=peer',
            'filter="(cn=*FooBar*)"',
            'searchbase="ou=äöüÄÖÜß"',
            'scope=one',
            'schemachecking=on',
            'type=refreshAndPersist',
            '',
            '',
        ]
        srd = SyncReplDesc('  '.join(srd_str_list))
        self.assertEqual(repr(srd), 'SyncReplDesc(rid=23)')
        self.assertEqual(srd.rid, '23')
        self.assertEqual(srd.provider, 'ldap://ldap-23.example.com')
        self.assertEqual(srd.bindmethod, 'simple')
        self.assertEqual(srd.timeout, 5)
        self.assertEqual(srd.network_timeout, 5)
        self.assertEqual(srd.saslmech, 'EXTERNAL')
        self.assertEqual(srd.keepalive, '240:10:30')
        self.assertEqual(srd.starttls, 'critical')
        self.assertEqual(srd.tls_cert, None)
        self.assertEqual(srd.tls_key, None)
        self.assertEqual(srd.tls_cacert, '/etc/ssl/ca-certs.pem')
        self.assertEqual(srd.tls_reqcert, 'try')
        self.assertEqual(srd.tls_protocol_min, None)
        self.assertEqual(srd.tls_crlcheck, 'peer')
        self.assertEqual(srd.filter, '(cn=*FooBar*)')
        self.assertEqual(srd.searchbase, 'ou=äöüÄÖÜß')
        self.assertEqual(srd.scope, SCOPE_ONELEVEL)
        self.assertEqual(srd.schemachecking, 'on')
        self.assertEqual(srd.type, 'refreshAndPersist')
        self.assertEqual(srd.retry, None)
        srd_url_obj = srd.ldap_url()
        self.assertEqual(srd_url_obj.dn, 'ou=äöüÄÖÜß')
        self.assertEqual(srd_url_obj.scope, SCOPE_ONELEVEL)
        self.assertEqual(
            str(srd_url_obj),
            'ldap://ldap-23.example.com/ou%3D%C3%A4%C3%B6%C3%BC%C3%84%C3%96%C3%9C%C3%9F?*,+?one?%28cn%3D%2AFooBar%2A%29'
        )

    def test003_value_error(self):
        for invalid_val in (
            'timeout=1x2',
            'bindmethod=foo',
            'scope=foo',
            'starttls=foo',
        ):
            try:
                SyncReplDesc(invalid_val)
            except ValueError:
                pass
            else:
                self.fail('%r should have raised ValueError' % (invalid_val))


class TestLdapSearchCmd(unittest.TestCase):
    """
    test ldap0.openldap.ldapsearch_cmd()
    """
    maxDiff = None

    def test001_ldapsearch_cmd(self):
        self.assertEqual(
            ldapsearch_cmd(
                LDAPUrl(
                    'ldap://127.0.0.1:1234/'
                    'dc=example,dc=com'
                    '?attr1,attr2,attr3'
                    '?sub'
                    '?%28objectClass%3D%2A%29'
                    '?'
                    'bindname=cn%3Dfred%2Cc%3Dau,'
                    'X-BINDPW=%3F%3F%3F,'
                    'trace=8,'
                    'x-saslmech=DIGEST-MD5,'
                    'x-saslrealm=example.com,'
                    'x-saslauthzid=u:anna,'
                    'x-starttls=2,'
                )
            ),
            ' '.join((
                'ldapsearch',
                '-LL',
                '-H "ldap://127.0.0.1:1234"',
                '-ZZ','-b "dc=example,dc=com"',
                '-s sub',
                '-Y "DIGEST-MD5"',
                '-X "u:anna"',
                '"(objectClass=*)"',
                'attr1 attr2 attr3',
            ))
        )
        self.assertEqual(
            ldapsearch_cmd(
                LDAPUrl(
                    'ldaps://ldap.example.com/'
                    'dc=example,dc=com'
                    '?'
                    '?one'
                    '?(uid=foo)'
                )
            ),
            'ldapsearch -LL -H "ldaps://ldap.example.com"  -b "dc=example,dc=com" -s one -x -D "" -w ""  "(uid=foo)"'
        )


if __name__ == '__main__':
    unittest.main()
